#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
curveprops: Model and view for curve properties
"""

#raise NotImplementedError('Under Construction!')

from __future__ import absolute_import
from builtins import str
from builtins import bytes
from builtins import object

import copy

from taurus.external.qt import Qt, Qwt5
import taurus
import taurus.core
from taurus.core import TaurusElementType
from taurus.qt.qtcore.mimetypes import TAURUS_MODEL_LIST_MIME_TYPE, TAURUS_ATTR_MIME_TYPE
from taurus.qt.qtgui.util.ui import UILoadable

from .curvesAppearanceChooserDlg import NamedLineStyles, ReverseNamedLineStyles, \
    NamedCurveStyles, ReverseNamedCurveStyles, \
    NamedSymbolStyles, ReverseNamedSymbolStyles, \
    NamedColors, CurveAppearanceProperties


__all__ = ['CurveConf', 'CurvesTableModel',
           'ExtendedSelectionModel', 'CurvePropertiesView']

# set some named constants
# columns:
NUMCOLS = 4
X, Y, TITLE, VIS = list(range(NUMCOLS))
SRC_ROLE = Qt.Qt.UserRole + 1
PROPS_ROLE = Qt.Qt.UserRole + 2

from taurus import isValidName

class Component(object):

    def __init__(self, src):
        self.src = src
        self.display = ''
        self.icon = Qt.QIcon()
        self.ok = True
        self._dbCache = taurus.Authority()
        self.setSrc(src)

    def update(self):
        self.setSrc(self.src)

    def setSrc(self, src):
        self.src, self.display, self.icon, self.ok = self.processSrc(src)

    def processSrc(self, src):
        '''returns src,display,icon,ok'''
        src = str(src)
        # empty
        if src == '':
            return '', '', Qt.QIcon(), True
        # for formulas
        if src.startswith('='):
            #@todo: evaluate/validate the expression
            return src, src[1:].strip(), Qt.QIcon.fromTheme('accessories-calculator'), True
        # for attributes
        if isValidName(src, etypes=[TaurusElementType.Attribute]):
            return src, src, Qt.QIcon.fromTheme('network-server'), True
        # If nothing matches...
        return src, src, Qt.QIcon.fromTheme('dialog-warning'), False


class CurveConf(object):

    def __init__(self, xsrc='', ysrc='', properties=None, title='', vis=Qwt5.QwtPlot.yLeft):
        if properties is None:
            properties = CurveAppearanceProperties()
        self.properties = properties
        self.title = title
        self.vis = vis
        self.x = Component(xsrc)
        self.y = Component(ysrc)

    def __repr__(self):
        ret = "CurveConf(xsrc='%s', ysrc='%s', title='%s')" % (
            self.x.src, self.y.src, self.title)
        return ret


class CurvesTableModel(Qt.QAbstractTableModel):
    ''' A model to manage information about curves to be plotted an their appearance
    '''

    dataChanged = Qt.pyqtSignal('QModelIndex', 'QModelIndex')

    def __init__(self, curves=None):
        if curves is None:
            curves = []
        super(CurvesTableModel, self).__init__()
        self.ncolumns = NUMCOLS
        self.curves = curves

    def addCurve(self, curveconf):
        self.curves.append(curveconf)

    def dumpData(self):
        #        return copy.deepcopy(self.curves)
        return copy.copy(self.curves)
#        return self.curves

    def rowCount(self, index=Qt.QModelIndex()):
        return len(self.curves)

    def columnCount(self, index=Qt.QModelIndex()):
        return self.ncolumns

    def data(self, index, role=Qt.Qt.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return None
        row = index.row()
        column = index.column()
        # Display Role
        if role == Qt.Qt.DisplayRole:
            if column == X:
                return str(self.curves[row].x.display)
            elif column == Y:
                return str(self.curves[row].y.display)
            elif column == TITLE:
                return str(self.curves[row].title)
            elif column == VIS:
                return str(self.curves[row].vis)
            else:
                return None
        elif role == Qt.Qt.DecorationRole:
            if column == X:
                return self.curves[row].x.icon
            elif column == Y:
                return self.curves[row].y.icon
            elif column == TITLE:
                return Qt.QColor(self.curves[row].properties.lColor or 'black')
            else:
                return None
        elif role == Qt.Qt.TextColorRole:
            if column == X:
                Qt.QColor(self.curves[row].x.ok and 'green' or 'red')
            elif column == Y:
                Qt.QColor(self.curves[row].y.ok and 'green' or 'red')
            else:
                return None
        elif role == SRC_ROLE:
            if column == X:
                return str(self.curves[row].x.src)
            elif column == Y:
                return str(self.curves[row].y.src)
            else:
                return None
        elif role == PROPS_ROLE:
            return self.curves[row].properties
        elif role == Qt.Qt.ToolTipRole:
            if column == X:
                return str(self.curves[row].x.src)
            elif column == Y:
                return str(self.curves[row].y.src)
            else:
                return None
        if role == Qt.Qt.EditRole:
            if column == X:
                return str(self.curves[row].x.src)
            elif column == Y:
                return str(self.curves[row].y.src)
            elif column == TITLE:
                return str(self.curves[row].title)
            else:
                return None
        # Alignment
#         elif role == Qt.Qt.TextAlignmentRole:
#             return int(Qt.AlignHCenter|Qt.AlignVCenter)
        # Text Color
#        elif role == Qt.Qt.TextColorRole:
#            return Qt.QColor(self.curves[row].properties.lColor or 'black')
        return None

    def headerData(self, section, orientation, role=Qt.Qt.DisplayRole):
        if role == Qt.Qt.TextAlignmentRole:
            if orientation == Qt.Qt.Horizontal:
                return int(Qt.Qt.AlignLeft | Qt.Qt.AlignVCenter)
            return int(Qt.Qt.AlignRight | Qt.Qt.AlignVCenter)
        if role != Qt.Qt.DisplayRole:
            return None
        # So this is DisplayRole...
        if orientation == Qt.Qt.Horizontal:
            if section == X:
                return "X source"
            elif section == Y:
                return "Y Source"
            elif section == TITLE:
                return "Title"
            elif section == VIS:
                return "Shown at"
            return None
        else:
            return str(section + 1)

    def flags(self, index):  # use this to set the editable flag when fix is selected
        if not index.isValid():
            return Qt.Qt.ItemIsEnabled
        column = index.column()
        if column in (X, Y):
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled | Qt.Qt.ItemIsDropEnabled | Qt.Qt.ItemIsSelectable)
        elif column == TITLE:
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsDragEnabled)
        return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsDragEnabled)

    def setData(self, index, value=None, role=Qt.Qt.EditRole):
        if index.isValid() and (0 <= index.row() < self.rowCount()):
            row = index.row()
            curve = self.curves[row]
            if role == PROPS_ROLE:
                self.curves[row].properties = value
                self.dataChanged.emit(self.index(
                    row, 0), self.index(row, self.ncolumns - 1))
            else:
                column = index.column()
                if column == X:
                    curve.x.setSrc(value)
                elif column == Y:
                    curve.y.setSrc(value)
                elif column == TITLE:
                    curve.title = value
                elif column == VIS:
                    curve.vis = value
                self.dataChanged.emit(index, index)
            return True
        return False

    def insertRows(self, position=None, rows=1, parentindex=None):
        if position is None:
            position = self.rowCount()
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginInsertRows(parentindex, position, position + rows - 1)
        slice = [CurveConf() for i in range(rows)]
        self.curves = self.curves[:position] + slice + self.curves[position:]
        self.endInsertRows()
        return True

    def removeRows(self, position, rows=1, parentindex=None):
        if parentindex is None:
            parentindex = Qt.QModelIndex()
        self.beginResetModel()
        self.beginRemoveRows(parentindex, position, position + rows - 1)
        self.curves = self.curves[:position] + self.curves[position + rows:]
        self.endRemoveRows()
        self.endResetModel()
        return True

    def mimeTypes(self):
        result = list(Qt.QAbstractTableModel.mimeTypes(self))
        result += [TAURUS_ATTR_MIME_TYPE, 'text/plain']
        return result

    def dropMimeData(self, data, action, row, column, parent):
        if row == -1:
            if parent.isValid():
                row = parent.row()
            else:
                row = parent.rowCount()
        if column == -1:
            if parent.isValid():
                column = parent.column()
            else:
                column = parent.columnCount()
        if data.hasFormat(TAURUS_ATTR_MIME_TYPE):
            model = bytes(data.data(TAURUS_ATTR_MIME_TYPE)).decode('utf-8')
            self.setData(self.index(row, column), value=model)
            return True
        elif data.hasFormat(TAURUS_MODEL_LIST_MIME_TYPE):
            d = bytes(data.data(TAURUS_MODEL_LIST_MIME_TYPE))
            models = d.decode('utf-8').split()
            if len(models) == 1:
                self.setData(self.index(row, column), value=models[0])
                return True
            else:
                self.insertRows(row, len(models))
                for i, m in enumerate(models):
                    self.setData(self.index(row + i, column), value=m)
                return True
        elif data.hasText():
            self.setData(self.index(row, column), data.text())
            return True
        return False

    def mimeData(self, indexes):
        mimedata = Qt.QAbstractTableModel.mimeData(self, indexes)
        if len(indexes) == 1:
            txt = self.data(indexes[0], role=SRC_ROLE)
            # mimedata.setData(TAURUS_ATTR_MIME_TYPE, txt)
            mimedata.setText(txt)
        return mimedata
        # mimedata.setData()


class ExtendedSelectionModel(Qt.QItemSelectionModel):
    '''A QItemSelectionModel subclass that provides :meth:`partiallySelectedRows`'''

    def partiallySelectedRows(self):
        '''
        Returns the row numbers of those rows for which at least one column is selected.

        :return: (list<QModelIndex>) a list of indexes corresponding to the
                 first column of the (partially) selected curves
        '''
        temp = []
        for index in self.selectedIndexes():
            row = index.row()
            if row not in temp:
                temp.append(row)

        model = self.model()
        return [model.index(row, 0) for row in temp]


@UILoadable(with_ui='ui')
class CurvePropertiesView(Qt.QAbstractItemView):
    '''This widget is a view on a CurvesTableModel. It displays and allows to change the
    properties of selected curve(s). Note that this widget does not allow to
    change selection by itself, but rather relies on some other view on the same
    model (like a QTableView) to alter the selection.
    '''

    def __init__(self, parent=None, designMode=False):
        super(CurvePropertiesView, self).__init__(parent)
        self.loadUi()

        self.ui.sStyleCB.insertItems(0, sorted(NamedSymbolStyles.values()))
        self.ui.lStyleCB.insertItems(0, list(NamedLineStyles.values()))
        self.ui.cStyleCB.insertItems(0, list(NamedCurveStyles.values()))
        self.ui.sColorCB.addItem("")
        self.ui.lColorCB.addItem("")
        for color in NamedColors:
            icon = self._colorIcon(color)
            self.ui.sColorCB.addItem(icon, "", Qt.QColor(color))
            self.ui.lColorCB.addItem(icon, "", Qt.QColor(color))

        self._emptyProps = CurveAppearanceProperties()
        self.showProperties(self._emptyProps)

        # Connections
        self.ui.sStyleCB.currentIndexChanged.connect(self._onSymbolStyleChanged)
        self.ui.sStyleCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.lStyleCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.lStyleCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.lColorCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.sColorCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.cStyleCB.currentIndexChanged.connect(self.onPropertyControlChanged)
        self.ui.sSizeSB.valueChanged.connect(self.onPropertyControlChanged)
        self.ui.lWidthSB.valueChanged.connect(self.onPropertyControlChanged)
        self.ui.sFillCB.stateChanged.connect(self.onPropertyControlChanged)
        self.ui.cFillCB.stateChanged.connect(self.onPropertyControlChanged)

    #-------------------------------------------------------------------------
    # Reimplemented functions from base class

    def dataChanged(self, topleft, bottomright):
        '''Reimplemented. See :meth:`Qt.QAbstractItemView.dataChanged` '''
        Qt.QAbstractItemView.dataChanged(self, topleft, bottomright)

        # update what the controls are showing only if the changed data affects
        # what is selected
        minrow = topleft.row()
        maxrow = bottomright.row()
        for index in self.selectionModel().partiallySelectedRows():
            row = index.row()
            if row > minrow and row < maxrow:
                self.updateControls()
                return

    def selectionChanged(self, selected, deselected):
        '''Reimplemented. See :meth:`Qt.QAbstractItemView.selectionChanged` '''
        Qt.QAbstractItemView.selectionChanged(self, selected, deselected)
        self.updateControls()

    def indexAt(self, *args, **kwargs):
        ''' dummy reimplementation'''
        return Qt.QModelIndex()

    def visualRect(self, *args, **kwargs):
        ''' dummy reimplementation'''
        return Qt.QRect()

    def verticalOffset(self, *args, **kwargs):
        ''' dummy reimplementation'''
        return 0

    def horizontalOffset(self, *args, **kwargs):
        ''' dummy reimplementation'''
        return 0

    def visualRegionForSelection(self, *args, **kwargs):
        ''' dummy reimplementation'''
        return Qt.QRegion()

    def scrollTo(self, *args, **kwargs):
        ''' dummy reimplementation'''
        pass

    #-------------------------------------------------------------------------

    def _onSymbolStyleChanged(self, text):
        '''Slot called when the Symbol style is changed, to ensure that symbols
        are visible if you choose them

        :param text: (str) the new symbol style label
        '''
        text = str(text)
        if self.ui.sSizeSB.value() < 2 and not text in ["", "No symbol"]:
            # a symbol size of 0 is invisible and 1 means you should use
            # cStyle=dots
            self.ui.sSizeSB.setValue(3)

    def blockControlsSignals(self, block):
        '''blocks/unblocks the signals from all the properties controls

        :param block: (bool) If True, signals are blocked. If False they are unblocked
        '''
        self.ui.sStyleCB.blockSignals(block)
        self.ui.lStyleCB.blockSignals(block)
        self.ui.lColorCB.blockSignals(block)
        self.ui.cStyleCB.blockSignals(block)
        self.ui.sSizeSB.blockSignals(block)
        self.ui.lWidthSB.blockSignals(block)
        self.ui.sFillCB.blockSignals(block)
        self.ui.cFillCB.blockSignals(block)
        self.ui.sColorCB.blockSignals(block)

    def updateControls(self):
        '''Updates the state of the properties controls to reflect the selection
        '''
        selectedCurves = self.selectionModel().partiallySelectedRows()
        self.showProperties(self._emptyProps)
        if len(selectedCurves) < 1:
            return
        modeldata = self.model().data
        props = [modeldata(index, PROPS_ROLE) for index in selectedCurves]
        mergedprop = CurveAppearanceProperties.merge(
            props, conflict=CurveAppearanceProperties.inConflict_none)
        self.showProperties(mergedprop)

    def onPropertyControlChanged(self, *args):
        '''slot called whenever one of the controls is changed '''
        shownprop = self.getShownProperties()
        data = self.model().data
        setData = self.model().setData
        for index in self.selectionModel().partiallySelectedRows():
            prop = data(index, PROPS_ROLE)
            if prop.conflictsWith(shownprop):
                updatedprop = CurveAppearanceProperties.merge(
                    [prop, shownprop], conflict=CurveAppearanceProperties.inConflict_update_a)
                setData(index, updatedprop, PROPS_ROLE)

    def showProperties(self, prop, blockSignals=True):
        '''Updates the control widgets to show the given properties.

         ..note:: that the signals of the controls may be temporally blocked to
                  prevent loops. See the `blockSignals` parameter.

        :param prop: (CurveAppearanceProperties) the properties object
                     containing what should be shown. If a given property is set
                     to None, the corresponding widget will show a "neutral"
                     display
        :param blockSignals: (bool) If True (default) the signals of the control widgets
                             are blocked while updating them to avoid loops.
        '''
        if blockSignals:
            self.blockControlsSignals(True)
        # set the Style comboboxes
        self.ui.sStyleCB.setCurrentIndex(
            self.ui.sStyleCB.findText(NamedSymbolStyles[prop.sStyle]))
        self.ui.lStyleCB.setCurrentIndex(
            self.ui.lStyleCB.findText(NamedLineStyles[prop.lStyle]))
        self.ui.cStyleCB.setCurrentIndex(
            self.ui.cStyleCB.findText(NamedCurveStyles[prop.cStyle]))
        # set sSize and lWidth spinboxes. if prop.sSize is None, it puts -1
        # (which is the special value for these switchhboxes)
        self.ui.sSizeSB.setValue(max(prop.sSize, -1))
        self.ui.lWidthSB.setValue(max(prop.lWidth, -1))
        # Set the Color combo boxes. The item at index 0 is the empty one in
        # the comboboxes Manage unknown colors by including them
        if prop.sColor is None:
            index = 0
        else:
            index = self.ui.sColorCB.findData(Qt.QColor(prop.sColor))
        if index == -1:  # if the color is not one of the supported colors, add it to the combobox
            index = self.ui.sColorCB.count()  # set the index to what will be the added one
            self.ui.sColorCB.addItem(self._colorIcon(Qt.QColor(prop.sColor)),
                                     "",
                                     Qt.QColor(prop.sColor)
                                     )
        self.ui.sColorCB.setCurrentIndex(index)
        if prop.lColor is None:
            index = 0
        else:
            index = self.ui.lColorCB.findData(Qt.QColor(prop.lColor))
        if index == -1:  # if the color is not one of the supported colors, add it to the combobox
            index = self.ui.lColorCB.count()  # set the index to what will be the added one
            self.ui.lColorCB.addItem(self._colorIcon(Qt.QColor(prop.lColor)),
                                     "",
                                     Qt.QColor(prop.lColor)
                                     )
        self.ui.lColorCB.setCurrentIndex(index)
        # set the Fill Checkbox. The prop.sFill value can be in 3 states: True,
        # False and None
        if prop.sFill is None:
            checkState = Qt.Qt.PartiallyChecked
        elif prop.sFill:
            checkState = Qt.Qt.Checked
        else:
            checkState = Qt.Qt.Unchecked
        self.ui.sFillCB.setCheckState(checkState)
        # set the Area Fill Checkbox. The prop.cFill value can be in 3 states:
        # True, False and None
        if prop.cFill is None:
            checkState = Qt.Qt.PartiallyChecked
        elif prop.cFill:
            checkState = Qt.Qt.Checked
        else:
            checkState = Qt.Qt.Unchecked
        self.ui.cFillCB.setCheckState(checkState)
        if blockSignals:
            self.blockControlsSignals(False)

    def getShownProperties(self):
        """Returns a copy of the currently shown properties

        :return: (CurveAppearanceProperties)
        """
        prop = CurveAppearanceProperties()
        # get the values from the Style comboboxes. Note that the empty string
        # ("") translates into None
        prop.sStyle = ReverseNamedSymbolStyles[
            str(self.ui.sStyleCB.currentText())]
        prop.lStyle = ReverseNamedLineStyles[
            str(self.ui.lStyleCB.currentText())]
        prop.cStyle = ReverseNamedCurveStyles[
            str(self.ui.cStyleCB.currentText())]
        # get sSize and lWidth from the spinboxes
        prop.sSize = self.ui.sSizeSB.value()
        prop.lWidth = self.ui.lWidthSB.value()
        if prop.sSize < 0:
            prop.sSize = None
        if prop.lWidth < 0:
            prop.lWidth = None
        # Get the Color combo boxes. The item at index 0 is the empty one in
        # the comboboxes
        index = self.ui.sColorCB.currentIndex()
        if index == 0:
            prop.sColor = None
        else:
            prop.sColor = Qt.QColor(self.ui.sColorCB.itemData(index))
        index = self.ui.lColorCB.currentIndex()
        if index == 0:
            prop.lColor = None
        else:
            prop.lColor = Qt.QColor(self.ui.lColorCB.itemData(index))
        # get the sFill from the Checkbox.
        checkState = self.ui.sFillCB.checkState()
        if checkState == Qt.Qt.PartiallyChecked:
            prop.sFill = None
        else:
            prop.sFill = bool(checkState)
        # get the cFill from the Checkbox.
        checkState = self.ui.cFillCB.checkState()
        if checkState == Qt.Qt.PartiallyChecked:
            prop.cFill = None
        else:
            prop.cFill = bool(checkState)
        # store the props
        return copy.deepcopy(prop)

    def _colorIcon(self, color, w=10, h=10):
        '''returns an icon consisting of a rectangle of the given color

        :param color: (QColor or something accepted by QColor creator) The color for the icon
        :param w: (int) width of the icon
        :param h: (int) height of the icon

        :return: (QIcon)
        '''
        # to do: create a border
        pixmap = Qt.QPixmap(w, h)
        pixmap.fill(Qt.QColor(color))
        return Qt.QIcon(pixmap)


def main():
    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(sys.argv, cmd_line_parser=None)

    curves = [CurveConf(xsrc='', ysrc='tango://host:1000/a/b/c/d', properties=None, title="tangocurve", vis=Qwt5.QwtPlot.yLeft),
              CurveConf(xsrc='=[1,2,3]', ysrc='=#2.x**2',
                        properties=None, title="parab", vis=Qwt5.QwtPlot.yLeft)
              ]
    form = CurvePropertiesView()
    form.setModel(CurvesTableModel(curves))
    form.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    import sys
    main()



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
This module provides several taurus CLI subcommands (plot, trend, image,...)
that show a widget which may have more than one alternative implementation.
The available registered implementations for each command can be listed with
the `--ls-alt` option. Specific implementations can be selected with the
`--use-alt` option.

These commands also honour the default implementation selection configured in
:mod:`taurus.tauruscustomsettings` via the `*_ALT` variables

The alternative implementations can be registered using the following
entry-point group names:

- `taurus.plot.alt`: the loaded object is expected to be a `TaurusPlot`
  implementation

- `taurus.trend.alt`: the loaded object is expected to be a `TaurusTrend`
  implementation

- `taurus.trend2d.alt`: the loaded object is expected to be a
  `TaurusTrend2dDialog` implementation

- `taurus.image.alt`: the loaded object is expected to be a `TaurusImageDialog`
  implementation


Expected API for alternative implementations
--------------------------------------------

When registering a widget class to be used for these commands, the classes need
to provide the following minimum API:

- For **all** the classes:
  - must be a Qt.QWidget
  - must implement `setModel()` and `getModel()` equivalent to those in
  :class:`taurus.qt.qtgui.base.TaurusBaseComponent`

- For TaurusPlot:
  - should implement `setXAxisMode(mode)` where mode is one of `"n"` or `"t"`
  - should implement `loadConfigFile(name)` equivalent to
  :meth:`taurus.qt.qtcore.configuration.BaseConfigurableClass.loadConfigFile`

- For TaurusTrend:
  - should implement `setXAxisMode(mode)` where mode is one of `"n"` or `"t"`
  - should implement  `loadConfigFile(name)` equivalent to
  :meth:`taurus.qt.qtcore.configuration.BaseConfigurableClass.loadConfigFile`
  - should implement `setMaxDataBufferSize(n)` where `n` is an integer
  - should implement `setForcedReadingPeriod(period)` where `period` is an
  integer

- For TaurusTrend2D:
  - should accept the following keyword arguments in its constructor:

    - stackMode=x_axis_mode where mode is one of `"d"` or `"n"` or `"t"`
    - wintitle=window_name,
    - buffersize=max_buffer_size

- For TaurusImage:
  - should accept the following keyword arguments in its constructor:

    - wintitle=window_name,

  - should implement `setRGBmode(mode)` where mode is one of `"gray"`, `"rgb"`

"""

import sys
import click
import taurus.cli.common
from taurus.core.util.plugin import selectEntryPoints
from taurus import tauruscustomsettings as _ts
from taurus import warning, info
from taurus.qt.qtgui.application import TaurusApplication


EP_GROUP_PLOT = "taurus.plot.alts"
EP_GROUP_TREND = "taurus.trend.alts"
EP_GROUP_TREND2D = "taurus.trend2d.alts"
EP_GROUP_IMAGE = "taurus.image.alts"


def _print_alts(group):
    alts = [ep.name for ep in selectEntryPoints(group)]
    print("Registered alternatives :\n {}\n".format("\n ".join(alts)))


def _load_class_from_group(group, include=(".*",), exclude=()):
    """
    Factory that returns the first available class from the group entry point.
    The selection is done among the classes registered in
    the `group` entry-point, prioritized according to the given
    `include` and `exclude` patterns
    (see :func:`taurus.core.util.plugin.selectEntryPoints`)
    """
    eps = selectEntryPoints(group, include=include, exclude=exclude)
    for ep in eps:
        try:
            return ep.load(), ep.name
        except:
            info("Cannot load %s", ep.name)
    raise ImportError("Could not load any class from {}".format(eps))


def x_axis_mode_option(choices=("t", "n")):
    hlp = {
        "n": "regular axis",
        "e": "event number",
        "t": "absolute time/date axis",
        "d": "delta time axis",
    }
    o = click.option(
        "-x",
        "--x-axis-mode",
        "x_axis_mode",
        type=click.Choice(choices),
        default=choices[0],
        show_default=True,
        help=(
            "X axis mode: "
            + ", ".join([k + " for " + hlp[k] for k in choices])
        ),
    )
    return o


def max_buffer_option(default):
    o = click.option(
        "-b",
        "--buffer",
        "max_buffer_size",
        type=int,
        default=default,
        show_default=True,
        help=(
            "Maximum number of values to be stacked "
            + "(when reached, the oldest values will be "
            + "discarded)"
        ),
    )
    return o


forced_read_option = click.option(
    "-r",
    "--forced-read",
    "forced_read_period",
    type=int,
    default=-1,
    metavar="MILLISECONDS",
    help="force re-reading of the attributes every MILLISECONDS ms",
)


@click.command("plot")
@taurus.cli.common.models
@taurus.cli.common.config_file
@x_axis_mode_option(["n", "t"])
@taurus.cli.common.demo
@taurus.cli.common.window_name("TaurusPlot")
@taurus.cli.common.use_alternative
@taurus.cli.common.list_alternatives
def plot_cmd(
    models, config_file, x_axis_mode, demo, window_name, use_alt, ls_alt
):
    """Shows a plot for the given models"""

    if ls_alt:
        _print_alts(EP_GROUP_PLOT)
        sys.exit(0)

    if use_alt is None:
        use_alt = getattr(_ts, "PLOT_ALT", ".*")

    try:
        TPlot, epname = _load_class_from_group(
            EP_GROUP_PLOT, include=[use_alt]
        )
    except:
        _print_alts(EP_GROUP_PLOT)
        sys.exit(1)

    app = TaurusApplication(app_name="taurusplot({})".format(epname))

    w = TPlot()
    w.setWindowTitle(window_name)

    if demo:
        models = list(models)
        models.extend(["eval:rand(100)", "eval:0.5*sqrt(arange(100))"])

    try:
        w.setXAxisMode(x_axis_mode)
    except Exception as e:
        warning(
            'Could not set X axis mode to "%s" on %s plot. Reason: "%s"',
            x_axis_mode,
            epname,
            e,
        )
        sys.exit(1)

    if config_file is not None:
        try:
            w.loadConfigFile(config_file)
        except Exception as e:
            warning(
                'Could not load config file "%s" on %s plot. Reason: "%s"',
                config_file,
                epname,
                e,
            )
            sys.exit(1)

    if models:
        w.setModel(models)

    w.show()
    sys.exit(app.exec_())


@click.command("trend")
@taurus.cli.common.models
@taurus.cli.common.config_file
@x_axis_mode_option(["t", "n"])
@taurus.cli.common.demo
@taurus.cli.common.window_name("TaurusTrend")
@taurus.cli.common.use_alternative
@taurus.cli.common.list_alternatives
@max_buffer_option(None)
@forced_read_option
def trend_cmd(
    models,
    config_file,
    x_axis_mode,
    demo,
    window_name,
    use_alt,
    ls_alt,
    max_buffer_size,
    forced_read_period,
):
    """Shows a trend for the given models"""

    # list alternatives option
    if ls_alt:
        _print_alts(EP_GROUP_TREND)
        sys.exit(0)

    # use alternative
    if use_alt is None:
        use_alt = getattr(_ts, "TREND_ALT", ".*")

    # get the selected alternative
    try:
        TTrend, epname = _load_class_from_group(
            EP_GROUP_TREND, include=[use_alt]
        )
    except:
        _print_alts(EP_GROUP_TREND)
        sys.exit(1)

    app = TaurusApplication(app_name="taurustrend({})".format(epname))
    w = TTrend()

    # window title option
    w.setWindowTitle(window_name)

    # demo option
    if demo:
        models = list(models)
        models.extend(["eval:rand()", "eval:1+rand(2)"])

    # x axis mode option
    try:
        w.setXAxisMode(x_axis_mode)
    except Exception as e:
        warning(
            'Could not set X axis mode to "%s" on %s plot. Reason: "%s"',
            x_axis_mode,
            epname,
            e,
        )
        sys.exit(1)

    # configuration file option
    if config_file is not None:
        try:
            w.loadConfigFile(config_file)
        except Exception as e:
            warning(
                'Could not load config file "%s" on %s plot. Reason: "%s"',
                config_file,
                epname,
                e,
            )
            sys.exit(1)

    # max buffer size option
    if max_buffer_size is not None:
        try:
            w.setMaxDataBufferSize(max_buffer_size)
        except Exception as e:
            warning(
                'Could not set max buffer size on %s trend. Reason: "%s"',
                epname,
                e,
            )
        sys.exit(1)

    # set models
    if models:
        w.setModel(list(models))

    # period option
    if forced_read_period > 0:
        w.setForcedReadingPeriod(forced_read_period)

    w.show()
    sys.exit(app.exec_())


@click.command("trend2d")
@taurus.cli.common.model
@x_axis_mode_option(["d", "t", "n", "e"])
@taurus.cli.common.demo
@taurus.cli.common.window_name("TaurusTrend2D")
@taurus.cli.common.use_alternative
@taurus.cli.common.list_alternatives
@max_buffer_option(512)
def trend2d_cmd(
    model, x_axis_mode, demo, window_name, use_alt, ls_alt, max_buffer_size
):

    # list alternatives option
    if ls_alt:
        _print_alts(EP_GROUP_TREND2D)
        sys.exit(0)

    # use alternative
    if use_alt is None:
        use_alt = getattr(_ts, "TREND2D_ALT", ".*")

    # get the selected alternative
    try:
        TTrend2D, epname = _load_class_from_group(
            EP_GROUP_TREND2D, include=[use_alt]
        )
    except:
        _print_alts(EP_GROUP_TREND2D)
        sys.exit(1)

    app = TaurusApplication(app_name="Taurus Trend 2D ({})".format(epname))
    w = TTrend2D(
        stackMode=x_axis_mode, wintitle=window_name, buffersize=max_buffer_size
    )

    if demo:
        model = "eval:x=linspace(0,3,40);t=rand();sin(x+t)"

    if model:
        w.setModel(model)

    w.show()
    sys.exit(app.exec_())


@click.command("image")
@taurus.cli.common.model
@taurus.cli.common.demo
@taurus.cli.common.window_name("TaurusImage")
@taurus.cli.common.use_alternative
@taurus.cli.common.list_alternatives
@click.option(
    "-c",
    "--color-mode",
    "color_mode",
    type=click.Choice(["gray", "rgb"]),
    default="gray",
    show_default=True,
    help=("Color mode expected from the attribute"),
)
def image_cmd(model, demo, window_name, color_mode, use_alt, ls_alt):
    # list alternatives option
    if ls_alt:
        _print_alts(EP_GROUP_IMAGE)
        sys.exit(0)

    # use alternative
    if use_alt is None:
        use_alt = getattr(_ts, "IMAGE_ALT", ".*")

    # get the selected alternative
    try:
        TImage, epname = _load_class_from_group(
            EP_GROUP_IMAGE, include=[use_alt]
        )
    except:
        _print_alts(EP_GROUP_IMAGE)
        sys.exit(1)

    app = TaurusApplication(app_name="Taurus Image ({})".format(epname))

    rgb_mode = color_mode == "rgb"

    # TODO:  is "-c rgb --demo" doing the right thing?? Check it.
    if demo:
        if color_mode == "rgb":
            model = "eval:randint(0,256,(10,20,3))"
        else:
            model = "eval:rand(256,128)"

    w = TImage(wintitle=window_name)

    w.setRGBmode(rgb_mode)

    # set model
    if model:
        w.setModel(model)

    w.show()
    sys.exit(app.exec_())



#!/usr/bin/env python
#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Epics module. See __init__.py for more detailed documentation
"""

from __future__ import print_function
from __future__ import absolute_import

import numpy
from taurus.core.units import Quantity
from taurus.core.taurusbasetypes import (TaurusEventType, TaurusAttrValue,
                                         TaurusTimeVal, AttrQuality, DataType,
                                         DataFormat)
from taurus.core.taurusattribute import TaurusAttribute
import epics
from epics.ca import ChannelAccessException
from epics import dbr


__all__ = ['EpicsAttribute']



# map for epics DBR types to Taurus Types.
Dbr2TaurusType = {dbr.STRING: DataType.String,
                  dbr.INT: DataType.Integer,
                  dbr.SHORT: DataType.Integer,
                  dbr.FLOAT: DataType.Float,
                  # dbr.ENUM: DataType.Integer,  # TODO: Support enum
                  dbr.CHAR: DataType.Bytes,
                  dbr.LONG: DataType.Integer,
                  dbr.DOUBLE: DataType.Float,
                  # TODO: if a time type is added to DataType, change the next
                  dbr.TIME_STRING: DataType.String,
                  dbr.TIME_INT: DataType.Integer,
                  dbr.TIME_SHORT: DataType.Integer,
                  dbr.TIME_FLOAT: DataType.Float,
                  # dbr.TIME_ENUM: DataType.Integer,  # TODO: Support enum
                  dbr.TIME_CHAR: DataType.Bytes,
                  dbr.TIME_LONG: DataType.Integer,
                  dbr.TIME_DOUBLE: DataType.Float,
                  dbr.CTRL_STRING: DataType.String,
                  dbr.CTRL_INT: DataType.Integer,
                  dbr.CTRL_SHORT: DataType.Integer,
                  dbr.CTRL_FLOAT: DataType.Float,
                  # dbr.CTRL_ENUM: DataType.Integer,  # TODO: Support enum
                  dbr.CTRL_CHAR: DataType.Bytes,
                  dbr.CTRL_LONG: DataType.Integer,
                  dbr.CTRL_DOUBLE: DataType.Float,
                  }


class EpicsAttribute(TaurusAttribute):
    """
    A :class:`TaurusAttribute` that gives access to an Epics Process Variable.

    .. seealso:: :mod:`taurus.core.epics`

    .. warning:: In most cases this class should not be instantiated directly.
                 Instead it should be done via the
                 :meth:`EpicsFactory.getAttribute`
    """
    # TODO: support non-numerical PVs

    def __init__(self, name='', parent=None, storeCallback=None):
        self.call__init__(TaurusAttribute, name, parent,
                          storeCallback=storeCallback)

        self._label = self.getSimpleName()
        self._value = None
        self.data_format = None
        self.type = None
        self._range = [None, None]
        self._alarm = [None, None]
        self._warning = [None, None]

        self.__pv = epics.PV(self.getNormalName(), callback=self.onEpicsEvent,
                             form='ctrl',
                             connection_callback=self.onEpicsConnectionEvent)
        self.__pv.wait_for_connection()

    def getPV(self):
        """Returns the underlying :obj:`epics.PV` object
        """
        return self.__pv

    def onEpicsEvent(self, **kwargs):
        """callback for PV changes"""
        # this is called from the ca thread.
        # TODO: Consider doing the following on a different thread
        self._value = self.decode(self.__pv)
        self.fireEvent(TaurusEventType.Change, self._value)

    def onEpicsConnectionEvent(self, **kwargs):
        """callback for PV connection changes"""
        if kwargs['conn']:
            self.debug('(re)connected to epics PV')
            if self._value is not None:
                self._value.error = None
        else:
            self.warning('Connection to epics PV lost')
            self._value.error = ChannelAccessException('PV "%s" not connected' %
                                                       kwargs['pvname'])
        self.fireEvent(TaurusEventType.Change, self._value)

    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # Necessary to overwrite from TaurusAttribute
    # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def encode(self, value):
        """encodes the value passed to the write method into
        a representation that can be written with :meth:`epics.PV.put`
        """
        if isinstance(value, Quantity):
            # convert to units of the PV
            value = value.to(self.__pv.units).magnitude
        return value

    def decode(self, pv):
        """Decodes an epics PV object into a TaurusValue, and also updates other
         properties of the Attribute object
        """
        attr_value = TaurusAttrValue()
        if not pv.connected:
            attr_value.error = ChannelAccessException('PV "%s" not connected' %
                                                      pv.pvname)
            return attr_value
        v = pv.value
        # type
        try:
            self.type = Dbr2TaurusType[pv.ftype]
        except KeyError:
            raise ValueError('Unsupported epics type "%s"' % pv.type)
        # writable
        self.writable = pv.write_access
        # data_format
        if numpy.isscalar(v):
            self.data_format = DataFormat._0D
        else:
            self.data_format = DataFormat(len(numpy.shape(v)))
        # units and limits support
        if self.type in (DataType.Integer, DataType.Float):
            v = Quantity(v, pv.units)
            self._range = self.__decode_limit(pv.lower_ctrl_limit,
                                              pv.upper_ctrl_limit)
            self._alarm = self.__decode_limit(pv.lower_alarm_limit,
                                              pv.upper_alarm_limit)
            self._warning = self.__decode_limit(pv.lower_warning_limit,
                                                pv.upper_warning_limit)

        # rvalue
        attr_value.rvalue = v
        # wvalue
        if pv.write_access:
            attr_value.wvalue = v
        # time
        if pv.timestamp is None:
            attr_value.time = TaurusTimeVal.now()
        else:
            attr_value.time = TaurusTimeVal.fromtimestamp(pv.timestamp)
        # quality
        if pv.severity > 0:
            attr_value.quality = AttrQuality.ATTR_ALARM
        else:
            attr_value.quality = AttrQuality.ATTR_VALID
        return attr_value

    def __decode_limit(self, l, h):
        units = self.__pv.units
        if l is None or numpy.isnan(l):
            l = None
        else:
            l = Quantity(l, units)
        if l is None or numpy.isnan(h):
            h = None
        else:
            h = Quantity(h, units)
        return [l, h]

    def write(self, value, with_read=True):
        value = self.encode(value)
        self.__pv.put(value, wait=True)
        if with_read:
            return self.read(cache=False)

    def read(self, cache=True):
        """returns the value of the attribute.

        :param cache: (bool) If True (default), the last calculated value will
                      be returned. If False, the referenced values will be re-
                      read and the transformation string will be re-evaluated

        :return: attribute value
        """
        if not cache:
            self.__pv.get(use_monitor=False)
            self._value = self.decode(self.__pv)
        return self._value

    def poll(self):
        v = self.read(cache=False)
        self.fireEvent(TaurusEventType.Periodic, v)

    def isUsingEvents(self):
        return True  # TODO: implement this
# ------------------------------------------------------------------------------

    def factory(self):
        from .epicsfactory import EpicsFactory
        return EpicsFactory()

    @classmethod
    def getNameValidator(cls):
        from .epicsvalidator import EpicsAttributeNameValidator
        return EpicsAttributeNameValidator()


if __name__ == '__main__':
    a = EpicsAttribute('ca:XXX:a', None)
    b = EpicsAttribute('ca:XXX:b', None)
    s = EpicsAttribute('ca:XXX:sum', None)

    a.write(3.)
    b.write(4.)
    s.read()

    print("!$!", s.read(cache=False))
    print("a,b,s", a.read().rvalue, b.read().rvalue, s.read().rvalue)
    print("DF=", a.getDataFormat(), DataFormat.whatis(a.getDataFormat()))



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from builtins import str

from taurus.external.qt import Qt
from taurus.core.units import Quantity

import numpy

import taurus.core
from taurus.core.taurusbasetypes import (DataFormat, DataType, TaurusEventType,
                                         TaurusElementType)
from taurus.qt.qtgui.util import PintValidator
from taurus.qt.qtgui.display import TaurusLabel
from taurus.qt.qtgui.container import TaurusWidget
from taurus.core.util.enumeration import Enumeration


__all__ = ["TaurusValuesTable"]

__docformat__ = 'restructuredtext'


def _value2Quantity(value, units):
    '''
    Creates a Quantity from value and forces units if the vaule is unitless

    :param value: (int, float or str) a number or a string from which a quantity
                  can be created
    :param units: (str or Pint units) Units to use if the value is unitless
    :return: (Quantity)
    '''
    q = Quantity(value)
    if q.unitless:
        q = Quantity(q, units)
    return q


class TaurusValuesIOTableModel(Qt.QAbstractTableModel):
    typeCastingMap = {'f': float, 'b': bool,
                      'u': int, 'i': int, 'S': str, 'U': str}
    # Need to have an array

    dataChanged = Qt.pyqtSignal('QModelIndex', 'QModelIndex')

    def __init__(self, size, parent=None):
        Qt.QAbstractTableModel.__init__(self, parent)
        self._parent = parent
        self._rtabledata = []
        self._wtabledata = []
        self._rowCount = size[0]
        self._columnCount = size[1]
        self._modifiedDict = {}
        self._attr = None
        self.editedIndex = None
        self._editable = False
        self._writeMode = False

    def isDirty(self):
        '''returns True if there are user changes. False Otherwise'''
        return bool(self._modifiedDict)

    # To be implemented -----
    def rowCount(self, index=Qt.QModelIndex()):
        '''see :meth:`Qt.QAbstractTableModel.rowCount`'''
        if self._rowCount == 0:
            self._rowCount = 1
        return self._rowCount

    def columnCount(self, index=Qt.QModelIndex()):
        '''see :meth:`Qt.QAbstractTableModel.columnCount`'''
        if self._columnCount == 0:
            self._columnCount = 1
        return self._columnCount

    def data(self, index, role=Qt.Qt.DisplayRole):
        '''see :meth:`Qt.QAbstractTableModel.data`'''
        if self._writeMode == False:
            tabledata = self._rtabledata
        else:
            tabledata = self._wtabledata
        if not index.isValid() or not (0 <= index.row() < len(tabledata)):
            return None
        elif role == Qt.Qt.DisplayRole:
            value = None
            rc = (index.row(), index.column())
            if self._writeMode and rc in self._modifiedDict:
                if self.getAttr().type in [DataType.Integer, DataType.Float]:
                    return str(self._modifiedDict[rc])
                else:
                    return self._modifiedDict[rc]
            else:
                value = tabledata[rc]
                if isinstance(value, Quantity):
                    value = value.magnitude
            # cast the value to a standard python type
            value = self.typeCastingMap[tabledata.dtype.kind](value)
            return value
        elif role == Qt.Qt.DecorationRole:
            if ((index.row(), index.column()) in self._modifiedDict
                    and self._writeMode):
                if self.getAttr().type in [DataType.Integer, DataType.Float]:
                    value = self._modifiedDict[(index.row(), index.column())]
                    if not self.inAlarmRange(value):
                        icon = Qt.QIcon.fromTheme('document-save')
                    else:
                        icon = Qt.QIcon.fromTheme('emblem-important')
                else:
                    icon = Qt.QIcon.fromTheme('document-save')
                return icon
        elif role == Qt.Qt.EditRole:
            value = None
            if ((index.row(), index.column()) in self._modifiedDict
                    and self._writeMode):
                value = self._modifiedDict[(index.row(), index.column())]
            else:
                value = tabledata[index.row(), index.column()]
                if tabledata.dtype == bool:
                    value = bool(value)
            return value
        elif role == Qt.Qt.BackgroundRole:
            if self._writeMode:
                return Qt.QColor(22, 223, 21, 50)
            else:
                return Qt.QColor('white')
        elif role == Qt.Qt.ForegroundRole:
            if ((index.row(), index.column()) in self._modifiedDict
                    and self._writeMode):
                if self.getAttr().type in [DataType.Integer, DataType.Float]:
                    value = self._modifiedDict[(index.row(), index.column())]
                    if not self.inAlarmRange(value):
                        return Qt.QColor('blue')
                    else:
                        return Qt.QColor('orange')
                else:
                    return Qt.QColor('blue')
            return Qt.QColor('black')
        elif role == Qt.Qt.FontRole:
            if ((index.row(), index.column()) in self._modifiedDict
                    and self._writeMode):
                return Qt.QFont("Arial", 10, Qt.QFont.Bold)
        elif role == Qt.Qt.ToolTipRole:
            if ((index.row(), index.column()) in self._modifiedDict
                    and self._writeMode):
                value = str(self._modifiedDict[(index.row(), index.column())])
                msg = 'Original value: %s.\nNew value that will be saved: %s' %\
                      (str(tabledata[index.row(), index.column()]), value)
                return msg
        return None

    def getAttr(self):
        return self._attr

    def setAttr(self, attr):
        '''
        Updated the internal table data from an attribute value

        :param attr: (DeviceAttribute)
        '''
        self._attr = attr
        rvalue = attr.rvalue
        if attr.type not in [DataType.Float, DataType.Integer]:
            rvalue = numpy.array(attr.rvalue)
        # reshape the table
        if attr.data_format == DataFormat._1D:
            rows, columns = len(rvalue), 1
        elif attr.data_format == DataFormat._2D:
            rows, columns = numpy.shape(rvalue)
        else:
            raise TypeError('Unsupported data format "%s"' %
                            repr(attr.data_format))

        if (self._rowCount != rows) or (self._columnCount != columns):
            self.beginResetModel()
            self.endResetModel()

        self._rowCount = rows
        self._columnCount = columns
        rvalue = rvalue.reshape(rows, columns)
        if attr.type in [DataType.Integer, DataType.Float]:
            units = self._parent.getCurrentUnits()
            rvalue = rvalue.to(units)
        self._rtabledata = rvalue
        self._editable = False
        self.dataChanged.emit(self.createIndex(0, 0), self.createIndex(rows - 1, columns - 1))

    def getStatus(self, index):
        '''
        Returns Status of the variable

        :returns:  (taurus.core.taurusbasetypes.AttrQuality)
        '''
        return self._attr.quality

    def getType(self):
        '''
        Returns the table data type.

        :returns:  (numpy.dtype)
        '''
        return self._rtabledata.dtype

    def addValue(self, index, value):
        '''adds a value to the dictionary of modified cell values

        :param index: (QModelIndex) table index
        :param value: (object)
        '''
        rtable_value = self._rtabledata[index.row()][index.column()]
        if self._attr.getType() in [DataType.Float, DataType.Integer]:
            units = self._parent.getCurrentUnits()
            value = _value2Quantity(value, units)
            equals = numpy.allclose(rtable_value, value.to(units))
        else:
            equals = bool(rtable_value == value)
        if not equals:
            self._modifiedDict[(index.row(), index.column())] = value
        else:
            self.removeValue(index)

    def removeValue(self, index):
        '''
        Removes index from dictionary

        :param index:  (QModelIndex) table index
        '''
        if (index.row(), index.column()) in self._modifiedDict:
            self._modifiedDict.pop((index.row(), index.column()))

    def flags(self, index):
        '''see :meth:`Qt.QAbstractTableModel`'''
        if not index.isValid():
            return Qt.Qt.ItemIsEnabled
        if self._editable:
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsEditable | Qt.Qt.ItemIsSelectable)
        else:
            return Qt.Qt.ItemFlags(Qt.Qt.ItemIsEnabled | Qt.Qt.ItemIsSelectable)

    def getModifiedWriteData(self):
        '''
        returns an array for the write data that includes the user modifications

        :return: (numpy.array) The write values including user modifications.
        '''
        table = self._wtabledata
        kind = table.dtype.kind
        if kind in 'SU':
            table = table.tolist()  # we want to allow the strings to be larger than the original ones
            for (r, c), v in self._modifiedDict.items():
                table[r][c] = v
            table = numpy.array(table, dtype=str)
        else:
            for k, v in self._modifiedDict.items():
                if kind in ['f', 'i', 'u']:
                    units = self._parent.getCurrentUnits()
                    q = _value2Quantity(v, units)
                    table[k] = q
                elif kind == 'b':
                    if str(v) == "true":
                        table[k] = True
                    else:
                        table[k] = False
                else:
                    raise TypeError('Unknown data type "%s"' % kind)
        # reshape if needed
        if self._attr.data_format == DataFormat._1D:
            table = table.flatten()
        return table

    def clearChanges(self):
        '''clears the dictionary of changed values'''
        self._modifiedDict.clear()
        self.dataChanged.emit(self.createIndex(
            0, 0), self.createIndex(self.rowCount() - 1, self.columnCount() - 1))

    def inAlarmRange(self, value):
        '''
        Checkes if value is in alarm range.

        :param value:  Quantity value
        :returns:  (bool) True if value in alarm range, False if valid

        '''
        try:
            min_alarm, max_alarm = self._attr.alarms
            if min_alarm >= value or value >= max_alarm:
                return True
            else:
                return False
        except:
            return True

    def inRange(self, value):
        '''
        Checks if value is in range.

        :param value:  Quantity value
        :returns:  (bool) True if value in range, False if valid

        '''
        try:
            min_range, max_range = self._attr.range
            if min_range <= value <= max_range:
                return True
            else:
                return False
        except:
            return True

    def setWriteMode(self, isWrite):
        '''Changes the write state

        :param isWrite: (bool)
        '''
        self._writeMode = isWrite
        if isWrite and not self.isDirty() and self._attr is not None:
            # refresh the write data (unless it is dirty)
            wvalue = self._attr.wvalue

            # reshape the table
            if self._attr.type == DataType.String:
                wvalue = numpy.array(wvalue)
            elif self._attr.type in [DataType.Integer, DataType.Float]:
                units = self._parent.getCurrentUnits()
                wvalue = wvalue.to(units)
            if self._attr.data_format == DataFormat._1D:
                rows, columns = numpy.shape(wvalue)[0], 1
                if rows == 0:
                    # TODO: Ask to the user for a default shape
                    rows = 3
            elif self._attr.data_format == DataFormat._2D:
                try:
                    rows, columns = numpy.shape(wvalue)
                except ValueError:
                    # TODO: Ask to the user for a default shape
                    rows = 3
                    columns = 3
                    wvalue = numpy.array((rows, columns))
            else:
                self.warning('unsupported data format %s' %
                             str(wvalue.data_format))
            wvalue = wvalue.reshape(rows, columns)
            # In version 4.6 of Qt when whole table is updated it is
            # recommended to use beginReset()
            self._wtabledata = wvalue
        self.dataChanged.emit(self.createIndex(0, 0), self.createIndex(self.rowCount() - 1,
                                                           self.columnCount() - 1))

    def getModifiedDict(self):
        '''
        Returns dictionary.

        :returns:  (dictionary) dictionary containing modified indexes and values
        '''
        return self._modifiedDict

    def getReadValue(self, index):
        '''
        Returns read value for a given index.

        :param index:  (QModelIndex) table model index
        :returns:  (string/int/float/bool) read table value for a given cell
        '''
        return self._rtabledata[index.row(), index.column()]


class TaurusValuesIOTable(Qt.QTableView):

    def __init__(self, parent=None):
        self._parent = parent
        name = self.__class__.__name__
        Qt.QTableView.__init__(self, parent)
        self._showQuality = True
        self._attr = None
        self._value = None
        self.setSelectionMode(Qt.QAbstractItemView.SingleSelection)
        itemDelegate = TaurusValuesIOTableDelegate(self)
        self.setItemDelegate(itemDelegate)
        #self.setStyleSheet('TaurusValuesIOTable { selection-background-color: violet;} ')

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # TaurusBaseWidget overwriting
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    def setModel(self, shape):
        '''
        Creates an instance of QTableModel and sets a QT model.

        :param shape:  (tuple<int>) shape of the model table to be set

        '''
        qmodel = TaurusValuesIOTableModel(shape, parent=self)
        Qt.QTableView.setModel(self, qmodel)

    def cancelChanges(self):
        '''
        Cancels all table modifications.
        '''
        self.model().clearChanges()

    def removeChange(self):
        '''
        If the cell was modified it restores the original write value.
        '''
        self.model().removeValue(self.selectedIndexes()[0])

    def showHelp(self):
        '''
        Shows QMessageBox help window. It contains explanations of used icons.
        '''
        buttonBox = Qt.QMessageBox(self)
        buttonBox.setLayout(Qt.QGridLayout(self))
        icon = Qt.QIcon.fromTheme('document-save').pixmap(48, 48)
        l = Qt.QLabel()
        l.setPixmap(icon)
        buttonBox.layout().addWidget(l, 0, 0)
        buttonBox.layout().addWidget(Qt.QLabel(
            '- value is valid and will be saved if changes will be accepted'), 0, 1)
        icon = Qt.QIcon.fromTheme('ddialog-warning').pixmap(48, 48)
        l = Qt.QLabel()
        l.setPixmap(icon)
        buttonBox.layout().addWidget(l, 1, 0)
        buttonBox.layout().addWidget(Qt.QLabel(
            '- value is in alarm range and will be saved if changes will be accepted'), 1, 1)
        buttonBox.exec_()

    def getCurrentUnits(self):
        try:
            return str(self._parent._units.currentText())
        except:
            return ''


class TaurusValuesIOTableDelegate(Qt.QStyledItemDelegate):

    editorCreated = Qt.pyqtSignal()

    def __init__(self, parent=None):
        Qt.QStyledItemDelegate.__init__(self, parent)
        self._parent = parent
        self._initialText = ""

    def createEditor(self, parent, option, index):
        '''
        Creates a custom editor for a table delagate.

        see :meth:`Qt.QStyledItemDelegate.createEditor`
        '''
        if index.model().getType() == bool:
            #editor = Qt.QStyledItemDelegate.createEditor(self,parent,option,index)
            editor = Qt.QComboBox(parent)
        else:
            editor = TableInlineEdit(parent)
            editor._updateValidator(index.model().getAttr())
        self.editorCreated.emit()
        return editor

    def setEditorData(self, editor, index):
        '''
        see :meth:`Qt.QStyledItemDelegate.setEditorData`
        '''
        if index.model().editedIndex == (index.row(), index.column()):
            return
        index.model().editedIndex = (index.row(), index.column())
        self._initialText = None
        if index.model().getType() == bool:
            editor.addItems(['true', 'false'])
            a = str(index.data()).lower()
            self._initialText = a

            editor.setCurrentIndex(editor.findText(a))
        else:
            data = index.model().data(index, Qt.Qt.EditRole)
            self._initialText = data
            editor.setText(str(self._initialText))

    def setModelData(self, editor, model, index):
        '''
        see :meth:`Qt.QStyledItemDelegate.setModelData`
        '''
        # if editor text changed, then don't mark as updated.
        isNumeric = False
        if self._parent._attr.type in [DataType.Integer, DataType.Float]:
            units = self._parent.getCurrentUnits()
            q = _value2Quantity(editor.text(), units)
            isNumeric = True
            if not model.inRange(q):
                return
        if index.model().getType() == bool:
            text = editor.currentText()
        else:
            if isNumeric:
                text = q
            else:
                text = editor.text()
            text = str(text)
        if(text != self._initialText) & (text != ""):
            model.addValue(index, text)
            hh = self.parent().horizontalHeader()
            if hh.length() > 0:
                try:
                    hh.setSectionResizeMode(hh.Fixed)
                except AttributeError:  # PyQt4
                    hh.setResizeMode(hh.Fixed)
            vh = self.parent().verticalHeader()
            if vh.length() > 0:
                try:
                    vh.setSectionResizeMode(vh.Fixed)
                except AttributeError:  # PyQt4
                    hh.setResizeMode(vh.Fixed)

        index.model().editedIndex = None


class TableInlineEdit(Qt.QLineEdit):
    '''
    TableInLineEdit is used to validate the content of the new value, also to paint the text: blue - valid, orange - in alarm, grey - invalid.
    '''

    def __init__(self, parent=None):
        super(Qt.QLineEdit, self).__init__(parent)
        self.textEdited.connect(self.onTextEdited)
        self.setValidator(None)
        self._min_range = None
        self._max_range = None
        self._min_alarm = None
        self._max_alarm = None
        self._default_unit = None

    def onTextEdited(self):
        """
        Paints the text while typing.

        Slot for the `Qt.QLineEdit.textEdited` signal
        """
        # default case: the value is in normal range with no pending changes
        color, weight = 'gray', 'normal'
        try:
            value = self.displayText()
            q = _value2Quantity(value, self._default_unit)
        except:
            q = 0.0
        try:
            if self._min_alarm < q < self._max_alarm:
                color = 'blue'
            elif self._min_range <= q <= self._max_range:
                # the value is valid but in alarm range...
                color = 'orange'
            else:
                # the value is invalid and can't be applied
                color = 'gray'
        except:
            color = 'gray'

        weight = 'bold'
        self.setStyleSheet('TableInlineEdit {color: %s; font-weight: %s}' %
                           (color, weight))

    def _updateValidator(self, attr):
        '''This method sets a validator depending on the data type

        :param attr: TaurusAttribute
        '''
        data_type = attr.getType()
        if data_type in [DataType.Integer, DataType.Float]:
            self._min_range, self._max_range = attr.range
            self._min_alarm, self._max_alarm = attr.alarms
            self._default_unit = attr.wvalue.units
            validator = PintValidator()
            validator.setBottom(self._min_range)
            validator.setTop(self._max_range)
            validator.setUnits(self._default_unit)
            self.setValidator(validator)
        else:
            self.setValidator(None)

    def __decimalDigits(self, fmt):
        '''returns the number of decimal digits from a format string
        (or None if they are not defined)'''
        try:
            if fmt[-1].lower() in ['f', 'g'] and '.' in fmt:
                return int(fmt[:-1].split('.')[-1])
            else:
                return None
        except:
            return None


class TaurusValuesTable(TaurusWidget):
    '''
    A table for displaying and/or editing 1D/2D Taurus attributes
    '''
    _showQuality = False
    _writeMode = False

    def __init__(self, parent=None, designMode=False,
                 defaultWriteMode=None):
        TaurusWidget.__init__(self, parent=parent, designMode=designMode)
        self._tableView = TaurusValuesIOTable(self)
        l = Qt.QGridLayout()
        l.addWidget(self._tableView, 1, 0)
        self._tableView.itemDelegate().editorCreated.connect(self._onEditorCreated)

        if defaultWriteMode is None:
            self.defaultWriteMode = "rw"
        else:
            self.defaultWriteMode = defaultWriteMode

        self._label = TaurusLabel()
        self._label.setBgRole('quality')
        self._label.setFgRole('quality')

        self._units = Qt.QComboBox()

        self._applyBT = Qt.QPushButton('Apply')
        self._cancelBT = Qt.QPushButton('Cancel')
        self._applyBT.clicked.connect(self.okClicked)
        self._cancelBT.clicked.connect(self.cancelClicked)

        self._rwModeCB = Qt.QCheckBox()
        self._rwModeCB.setText('Write mode')
        self._rwModeCB.toggled.connect(self.setWriteMode)

        lv = Qt.QHBoxLayout()
        lv.addWidget(self._label)
        lv.addWidget(self._units)
        l.addLayout(lv, 2, 0)
        l.addWidget(self._rwModeCB, 0, 0)
        lv = Qt.QHBoxLayout()
        lv.addWidget(self._applyBT)
        lv.addWidget(self._cancelBT)
        l.addLayout(lv, 3, 0)
        self._writeMode = False
        self.setLayout(l)
        self._initActions()

    def _initActions(self):
        """Initializes the actions for this widget (currently, the pause action.)
        """
        self._pauseAction = Qt.QAction("&Pause", self)
        self._pauseAction.setShortcuts([Qt.Qt.Key_P, Qt.Qt.Key_Pause])
        self._pauseAction.setCheckable(True)
        self._pauseAction.setChecked(False)
        self.addAction(self._pauseAction)
        self._pauseAction.toggled.connect(self.setPaused)
        self.chooseModelAction = Qt.QAction("Choose &Model", self)
        self.chooseModelAction.setEnabled(self.isModifiableByUser())
        self.addAction(self.chooseModelAction)
        self.chooseModelAction.triggered.connect(self.chooseModel)

    def getModelClass(self):
        '''see :meth:`TaurusWidget.getModelClass`'''
        return taurus.core.taurusattribute.TaurusAttribute

    def setModel(self, model):
        '''Reimplemented from :meth:`TaurusWidget.setModel`'''
        TaurusWidget.setModel(self, model)
        model_obj = self.getModelObj()

        if model_obj.isWritable() and self.defaultWriteMode != "r":
            self._writeMode = True
        else:
            self.defaultWriteMode = "r"

        if model_obj is not None:
            self._tableView._attr = model_obj
            if model_obj.type in [DataType.Integer, DataType.Float]:
                if self._writeMode:
                    try:
                        default_unit = str(model_obj.wvalue.units)
                    except AttributeError:
                        default_unit = ''
                else:
                    default_unit = str(model_obj.rvalue.units)
                # TODO: fill the combobox with the compatible units
                self._units.addItem("%s" % default_unit)
                self._units.setCurrentIndex(self._units.findText(default_unit))
                self._units.setEnabled(False)
            else:
                self._units.setVisible(False)
            raiseException = False
            if model_obj.data_format == DataFormat._2D:
                try:
                    dim_x, dim_y = numpy.shape(model_obj.rvalue)
                except ValueError:
                    raiseException = True
            elif model_obj.data_format == DataFormat._1D:
                try:
                    dim_x, dim_y = len(model_obj.rvalue), 1
                except ValueError:
                    raiseException = True
            else:
                raiseException = True
            if raiseException:
                raise Exception('rvalue is invalid')
            self._tableView.setModel([dim_x, dim_y])
        self.setWriteMode(self._writeMode)
        self._label.setModel(model)

    def handleEvent(self, evt_src, evt_type, evt_value):
        '''see :meth:`TaurusWidget.handleEvent`'''
        #@fixme: in some situations, we may miss some config event because of the qmodel not being set. The whole handleEvent Method and setModel method should be re-thought
        model = self._tableView.model()
        if model is None:
            return
        if evt_type in (TaurusEventType.Change,
                        TaurusEventType.Periodic)\
                and evt_value is not None:
            attr = self.getModelObj()
            model.setAttr(attr)
            model.setWriteMode(self._writeMode)

            hh = self._tableView.horizontalHeader()
            if hh.length() > 0:
                try:
                    hh.setSectionResizeMode(hh.Fixed)
                except AttributeError:  # PyQt4
                    hh.setResizeMode(hh.Fixed)
            vh = self._tableView.verticalHeader()
            if vh.length() > 0:
                try:
                    vh.setSectionResizeMode(vh.Fixed)
                except AttributeError:  # PyQt4
                    hh.setResizeMode(vh.Fixed)
            if self.defaultWriteMode == "r":
                isWritable = False
            else:
                isWritable = True
            writable = isWritable and self._writeMode and\
                attr.isWritable()
            self.setWriteMode(writable)
        elif evt_type == TaurusEventType.Config:
            # force a read to set an attr
            attr = self.getModelObj()
            model.setAttr(attr)

    def contextMenuEvent(self, event):
        '''Reimplemented from :meth:`QWidget.contextMenuEvent`'''
        menu = Qt.QMenu()
        globalPos = event.globalPos()
        menu.addAction(self.chooseModelAction)
        menu.addAction(self._pauseAction)
        if self._writeMode:
            index = self._tableView.selectedIndexes()[0]
            if index.isValid():
                val = self._tableView.model().getReadValue(index)
                if (index.row(), index.column()) in self._tableView.model().getModifiedDict():
                    menu.addAction(Qt.QIcon.fromTheme(
                        'edit-undo'), "Reset to original value (%s) " % repr(val), self._tableView.removeChange)
                    menu.addSeparator()
                menu.addAction(Qt.QIcon.fromTheme('process-stop'),
                               "Reset all table", self.askCancel)
                menu.addSeparator()
                menu.addAction(Qt.QIcon.fromTheme('help-browser'),
                               "Help", self._tableView.showHelp)
        menu.exec_(globalPos)
        event.accept()

    def applyChanges(self):
        '''
        Writes table modifications to the device server.
        '''
        tab = self._tableView.model().getModifiedWriteData()
        attr = self.getModelObj()
        if attr.type == DataType.String:
            # String arrays has to be converted to a list
            tab = tab.tolist()
        attr.write(tab)
        self._tableView.model().clearChanges()

    def okClicked(self):
        """This is a SLOT that is being triggered when ACCEPT button is clicked.

        .. note:: This SLOT is called, when user wants to apply table modifications.
                  When no cell was modified it will not be called. When
                  modifications have been done, they will be writen to w_value
                  of an attribute.
        """
        if self._tableView.model().isDirty():
            self.applyChanges()
            self.resetWriteMode()

    def cancelClicked(self):
        """This is a SLOT that is being triggered when CANCEL button is clicked.

        .. note:: This SLOT is called, when user does not want to apply table
                  modifications. When no cell was modified it will not be called.
        """
        if self._tableView.model().isDirty():
            self.askCancel()

    def askCancel(self):
        '''
        Shows a QMessageBox, asking if user wants to cancel all changes. Triggered when user clicks Cancel button.
        '''
        result = Qt.QMessageBox.warning(self, 'Your changes will be lost!',
                                        'Do you want to cancel changes done to the whole table?',
                                        Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel)
        if result == Qt.QMessageBox.Ok:
            self._tableView.cancelChanges()
            self.resetWriteMode()

    def _onEditorCreated(self):
        '''slot called when an editor has been created'''
        self.setWriteMode(self._writeMode)

    def getWriteMode(self):
        '''whether the widget is showing the read or write values

        :return: (bool)'''
        return self._writeMode

    def setWriteMode(self, isWrite):
        '''
        Triggered when the read mode is changed to write mode.

        :param isWrite: (bool)
        '''
        self._applyBT.setVisible(isWrite)
        self._cancelBT.setVisible(isWrite)
        self._rwModeCB.setChecked(isWrite)
        if self.defaultWriteMode in ("rw", "wr"):
            self._rwModeCB.setVisible(True)
        else:
            self._rwModeCB.setVisible(False)

        table_view_model = self._tableView.model()
        if table_view_model is not None:
            table_view_model.setWriteMode(isWrite)
            table_view_model._editable = isWrite
        if isWrite == self._writeMode:
            return
        self._writeMode = isWrite
        valueObj = self.getModelValueObj()
        if isWrite and valueObj is not None:
            w_value = valueObj.wvalue
            value = valueObj.rvalue
            if numpy.array(w_value).shape != numpy.array(value).shape:
                ta = self.getModelObj()
                v = ta.read()
                # @fixme: this is ugly! we should not be writing into the attribute without asking first...
                ta.write(v.rvalue)

    def resetWriteMode(self):
        '''equivalent to self.setWriteMode(self.defaultWriteMode)'''
        if self.defaultWriteMode == "r":
            isWritable = False
        else:
            isWritable = True
        self.setWriteMode(isWritable)

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''Reimplemented from :meth:`TaurusWidget.getQtDesignerPluginInfo`'''
        ret = TaurusWidget.getQtDesignerPluginInfo()
        ret['module'] = 'taurus.qt.qtgui.table'
        ret['group'] = 'Taurus Views'
        ret['icon'] = "designer:table.png"
        return ret

    def chooseModel(self):
        '''shows a model chooser'''
        from taurus.qt.qtgui.panel import TaurusModelChooser
        selectables = [TaurusElementType.Attribute]
        models, ok = TaurusModelChooser.modelChooserDlg(
            selectables=selectables, singleModel=True)
        if ok and len(models) == 1:
            self.setModel(models[0])

    def setModifiableByUser(self, modifiable):
        '''Reimplemented from :meth:`TaurusWidget.setModifiableByUser`'''
        self.chooseModelAction.setEnabled(modifiable)
        TaurusWidget.setModifiableByUser(self, modifiable)

    def isReadOnly(self):
        '''Reimplemented from :meth:`TaurusWidget.isReadOnly`'''
        return False

    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
    # QT property definition
    #-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    model = Qt.pyqtProperty("QString", TaurusWidget.getModel,
                            setModel,
                            TaurusWidget.resetModel)
    writeMode = Qt.pyqtProperty(
        "bool", getWriteMode, setWriteMode, resetWriteMode)


def taurusTableMain():
    '''A launcher for TaurusValuesTable.'''

    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.core.util import argparse
    import sys
    import os

    parser = argparse.get_taurus_parser()
    parser.set_usage("%prog [options] [model]]")
    parser.set_description(
        "A table for viewing and editing 1D and 2D attribute values")
    app = TaurusApplication(cmd_line_parser=parser,
                            app_name="TaurusValuesTable",
                            app_version=taurus.Release.version)
    args = app.get_command_line_args()

    dialog = TaurusValuesTable()
    dialog.setModifiableByUser(True)
    dialog.setWindowTitle(app.applicationName())

    # set a model list from the command line or launch the chooser
    if len(args) == 1:
        model = args[0]
        dialog.setModel(model)
    else:
        dialog.chooseModel()
        #model = 'sys/tg_test/1/boolean_spectrum'
        # model = 'sys/tg_test/1/boolean_image'
        # model = 'sys/tg_test/1/string_spectrum'
        # model = 'sys/tg_test/1/float_image'
        # model = 'sys/tg_test/1/double_image'
        # model = 'sys/tg_test/1/double_image_ro'
        # model = 'sys/tg_test/1/wave'
        # dialog.setModel(model)

    dialog.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    taurusTableMain()



import sys
from taurus.external.qt import Qt
from taurus.qt.qtgui.application import TaurusApplication

app = TaurusApplication(sys.argv, cmd_line_parser=None)
panel = Qt.QWidget()
layout = Qt.QHBoxLayout()
panel.setLayout(layout)

from taurus.qt.qtgui.display import TaurusLabel
w = TaurusLabel()
layout.addWidget(w)
w.model = 'sys/tg_test/1/double_scalar#rvalue.magnitude'

panel.show()
sys.exit(app.exec_())



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

# TODO: This module should be updated to use argparse instead of optparse

"""Helper command line parser for taurus based on :mod:`optparse`.
Suppose you have an application file::

    import sys
    from PyQt4 import Qt


    class GUI(Qt.QMainWindow):
        pass


    def main():
        import taurus.core.util.argparse as argparse

        parser, options, args = argparse.init_taurus_args()

        app = Qt.QApplication(sys.argv)
        w = GUI()
        w.show()
        sys.exit(app.exec_())

    if __name__ == "__main__":
        main()

The call to :func:`taurus.core.util.argparse.init_taurus_args` will initialize
taurus environment based on the command line options given by the user.
Currently, the known options are:

    #. ``--help`` prints the total number of available options
    #. ``--taurus-log-level`` sets the taurus log level
    #. ``--tango-host`` sets the default tango host
    #. ``--taurus-polling-period`` sets the default taurus global polling period (milliseconds)
    #. ``--taurus-serialization-mode`` sets the default taurus serialization mode
    #. ``--remote-console-port`` enables remote debugging using the given port

You can easily extend the taurus options with your application specific options.
Suppose you want to add an option like ``--model=<model name>``::

    def main():
        import taurus.core.util.argparse as argparse
        parser = argparse.get_taurus_parser(parser=parser)
        parser.set_usage("%prog [options] <special item>")
        parser.set_description("my own GUI application")
        parser.add_option("--model")

        parser, options, args = argparse.init_taurus_args(parser=parser)

        app = Qt.QApplication(sys.argv)
        w = GUI()
        w.show()
        sys.exit(app.exec_())
"""

__all__ = ["get_taurus_parser", "init_taurus_args", "parse_taurus_args",
           "split_taurus_args"]

__docformat__ = "restructuredtext"


from taurus.core.util import log as __log

__log.deprecated(dep='taurus.core.util.argparse', rel='4.5.4',
                 alt='argparse or (better) click')


def get_taurus_parser(parser=None):
    """ Returns a :class:`optparse.OptionParser` initialized with a
    :class:`optparse.OptionGroup` containning some taurus options.
    If a parser is given as parameter then it uses this parser instead of
    creating a new one.

    :param parser: an option parser or None (default) to create a new parser
    :type parser: :class:`optparse.OptionParser`
    :return: an option parser or the given parser if it is not None
    :rtype: :class:`optparse.OptionParser`"""
    import optparse
    if parser is None:
        parser = optparse.OptionParser()

    g = parser.get_option_group('--taurus-log-level')
    if g is None:
        group = optparse.OptionGroup(parser, "Taurus Options",
                                     "Basic options present in any taurus application")

        help_tauruslog = "taurus log level. Allowed values are (case insensitive): critical, "\
                         "error, warning, info, debug, trace"
        help_tangohost = "Tango host name (either HOST:PORT or a Taurus URI, e.g. tango://foo:1234)"
        help_tauruspolling = "taurus global polling period in milliseconds"
        help_taurusserial = "taurus serialization mode. Allowed values are (case insensitive): "\
            "serial, concurrent (default)"
        help_rcport = "enables remote debugging using the given port"
        help_formatter = "Override the default formatter"
        group.add_option("--taurus-log-level", dest="taurus_log_level", metavar="LEVEL",
                         help=help_tauruslog, type="str", default="info")
        group.add_option("--taurus-polling-period", dest="taurus_polling_period", metavar="MILLISEC",
                         help=help_tauruspolling, type="int", default=None)
        group.add_option("--taurus-serialization-mode", dest="taurus_serialization_mode", metavar="SERIAL",
                         help=help_taurusserial, type="str", default="Concurrent")
        group.add_option("--tango-host", dest="tango_host", metavar="TANGO_HOST",
                         help=help_tangohost, type="str", default=None)
        group.add_option("--remote-console-port", dest="remote_console_port", metavar="PORT",
                         help=help_rcport, type="int", default=None)
        group.add_option("--default-formatter", dest="default_formatter",
                         metavar="FORMATTER", help=help_formatter, type="str",
                         default=None)
        parser.add_option_group(group)
    return parser


def parse_taurus_args(parser=None, args=None, values=None):
    """Parses the command line. If parser is not given, then a new parser
    is created. In any case, the parser is initialized using the
    :func:`taurus.core.util.argparse.get_taurus_parser`.
    args and values are the optional parameters that will be given when
    executing :meth:`optparse.OptionParser.parse_args`.

    :param parser: an option parser or None (default) to create a new parser
    :type parser: :class:`optparse.OptionParser`
    :param args: the list of arguments to process (default is None meaning: sys.argv[1:])
    :type args: seq<str>
    :param values: a :class:`optparse.Values` object to store option arguments in
                  (default is None meaning: a new instance of Values) - if you give an
                  existing object, the option defaults will not be initialized on it
    :return: a tuple of three elements: parser, options, args
    :rtype: :class:`optparse.OptionParser`, :class:`optparse.Values`, seq<str> """
    parser = get_taurus_parser(parser=parser)
    options, args = parser.parse_args(args=args, values=values)
    return parser, options, args


def init_taurus_args(parser=None, args=None, values=None):
    """Parses the command line using :func:`taurus.core.util.argparse.parse_taurus_args`.

    After the command line is parsed, actions are taken on each recognized parameter.
    For example, the taurus log level and the default tango host are set accordingly.

    :param parser: an option parser or None (default) to create a new parser
    :type parser: :class:`optparse.OptionParser`
    :param args: the list of arguments to process (default is None meaning: sys.argv[1:])
    :type args: seq<str>
    :param values: a :class:`optparse.Values` object to store option arguments in
                  (default is None meaning: a new instance of Values) - if you give an
                  existing object, the option defaults will not be initialized on it
    :return: a tuple of three elements: parser, options, args
    :rtype: :class:`optparse.OptionParser`, :class:`optparse.Values`, seq<str> """
    import taurus
    parser, options, args = parse_taurus_args(
        parser=parser, args=args, values=values)

    # initialize taurus log level
    log_level_str = options.taurus_log_level.capitalize()
    if hasattr(taurus, log_level_str):
        log_level = getattr(taurus, log_level_str)
        taurus.setLogLevel(log_level)

    # initialize tango host
    if options.tango_host is not None:
        tango_factory = taurus.Factory("tango")
        tango_host = options.tango_host
        tango_factory.set_default_tango_host(tango_host)

    # initialize taurus polling period
    if options.taurus_polling_period is not None:
        taurus.Manager().changeDefaultPollingPeriod(options.taurus_polling_period)

    # initialize taurus serialization mode
    if options.taurus_serialization_mode is not None:
        import taurus.core.taurusbasetypes
        SerMode = taurus.core.taurusbasetypes.TaurusSerializationMode
        m = options.taurus_serialization_mode.capitalize()
        if hasattr(SerMode, m):
            m = getattr(SerMode, m)
            taurus.Manager().setSerializationMode(m)

    # initialize remote console port
    if options.remote_console_port is not None:
        try:
            import rfoo.utils.rconsole
            rfoo.utils.rconsole.spawn_server(port=options.remote_console_port)
            taurus.info("rconsole started. You can connect to it by typing: rconsole -p %d",
                        options.remote_console_port)
        except Exception as e:
            taurus.warning("Cannot spawn debugger. Reason: %s", e)

    # initialize default formatter
    if options.default_formatter is not None:
        from taurus import tauruscustomsettings
        setattr(tauruscustomsettings, 'DEFAULT_FORMATTER',
                options.default_formatter)

    return parser, options, args


def split_taurus_args(parser, args=None):
    """Splits arguments into valid parser arguments and non valid parser
    arguments.

    :param parser: an option parser
    :type parser: :class:`optparse.OptionParser`
    :param args: the list of arguments to process
                 (default is None meaning: sys.argv)
    :type args: seq<str>
    :return: a tuple of two elements: parser args, non parser args
    :rtype: seq<seq<str>, seq<str>>"""

    if args is None:
        import sys
        args = sys.argv

    taurus_args, non_taurus_args = [args[0]], [args[0]]
    for arg in args[1:]:
        arg_name = arg.split("=", 1)[0]
        if parser.has_option(arg_name):
            taurus_args.append(arg)
        else:
            non_taurus_args.append(arg)

    return taurus_args, non_taurus_args



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""The img submodule. It contains specific device implementation for CCDs and
2D detectors"""
from future.utils import string_types

from taurus.core.taurusbasetypes import TaurusEventType
from taurus.core.tango import TangoDevice
from taurus.core.util.containers import CaselessDict, CaselessList
from threading import Lock

__all__ = ['ImageDevice', 'ImageCounterDevice', 'PyImageViewer', 'ImgGrabber',
           'CCDPVCAM', 'ImgBeamAnalyzer', 'Falcon', 'LimaCCDs']

__docformat__ = 'restructuredtext'


class ImageDevice(TangoDevice):

    """A class encapsulating a generic image device"""

    def __init__(self, name, image_name='image', **kw):
        self.call__init__(TangoDevice, name, **kw)
        self.setImageAttrName(image_name)

    def addImageAttrName(self, attr_name):
        if attr_name in self._image_attr_names:
            return
        self._image_attr_names.append(attr_name)

    def setImageAttrName(self, attr_name):
        self._image_attr_names = CaselessList()
        self.addImageAttrName(attr_name)

    def getImageAttrName(self, idx=0):
        return self._image_attr_names[0]

    def getImageAttrNames(self):
        return self._image_attr_names


class ImageCounterDevice(ImageDevice):
    """A class encapsulating a generic image device that has an image counter
    attribute"""

    def __init__(self, name, image_name='image', image_ct='imagecounter',
                 **kw):
        self.lock = Lock()
        self._image_data = CaselessDict()
        self._image_id_attr_name = image_ct
        self.call__init__(ImageDevice, name, image_name, **kw)
        self.image_attr = self.getAttribute(image_name)
        self._image_id_attr = self.getAttribute(self._image_id_attr_name)
        self.discard_event = False
        self._image_id_attr.addListener(self)

    def _setDirty(self, names=None):
        names = names or self.getImageAttrNames()
        for n in names:
            d = self._image_data.get(n, (True, None))
            self._image_data[n] = True, d[1]

    def _getDirty(self, names=None):
        names = names or self.getImageAttrNames()
        dirty = []
        for name in names:
            d = self._image_data.get(name)
            if d is None or d[0] is True:
                dirty.append(name)
        return names

    def getImageIDAttrName(self):
        return self._image_id_attr_name

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_src == self._image_id_attr:
            if evt_type == TaurusEventType.Change:

                # discard events if there is one being processed
                if self.lock.locked():
                    self.discard_event = True
                    self.debug("Discard event %d" % evt_value.value)
                    return

                with self.lock:
                    while True:
                        self.debug("Processing image %d" % evt_value.rvalue)
                        # read the related Image attributes
                        # (asap and in one action)
                        images = self.getImageData()
                        self._setDirty()
                        self.fireEvent(evt_type, evt_value)
                        # maintain this fireEvent for backwards compatibility
                        # with Qub widget
                        self._emitImageEvents(evt_type, images)

                        if self.discard_event:
                            self.discard_event = False
                        else:
                            break
        else:
            ImageDevice.eventReceived(self, evt_src, evt_type, evt_value)

    def _emitImageEvents(self, evt_type, images):
        for attr_image_name in images:
            image_value = images[attr_image_name][1]
            if hasattr(image_value, 'is_empty') and not image_value.is_empty:
                self.debug("fireEvent for %s attribute" % attr_image_name)
                if not hasattr(image_value, 'rvalue'):
                    image_value.rvalue = image_value.value
                # Only emit to upper layers the events where
                # something has been read.
                attr_image = self.getAttribute(image_value.name)
                attr_image.fireEvent(evt_type, image_value)

    def getImageData(self, names=None):
        if names is None:
            names = self.getImageAttrNames()
        elif isinstance(names, string_types):
            names = (names,)

        fetch = self._getDirty(names)

        try:
            data = self.read_attributes(fetch)
            for d in data:
                self._image_data[d.name] = False, d
        except:
            pass
        return self._image_data


PyImageViewer = ImageCounterDevice
ImgGrabber = ImageCounterDevice
CCDPVCAM = ImageCounterDevice


class Falcon(ImageCounterDevice):

    def __init__(self, name, image_name='image', **kw):
        self._color = False
        self.call__init__(ImageCounterDevice, name,
                          image_name=image_name, **kw)
        self.imgFormat_Attr = self.getAttribute("imageformat")
        self.imgFormat_Attr.addListener(self)

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_src == self.getAttribute("imageformat"):
            if evt_type in (TaurusEventType.Change, TaurusEventType.Periodic):
                self._color = evt_value.value.lower() == "rgb24"
                return
        ImageCounterDevice.eventReceived(self, evt_src, evt_type, evt_value)

    def getImageData(self, names=None):
        data = ImageCounterDevice.getImageData(self, names=names)
        if self._color:
            for k, v in list(data.items()):
                s = v[1].value.shape
                v[1].value = v[1].value.reshape((s[0], s[1] // 3, 3))
        return data


class ImgBeamAnalyzer(ImageCounterDevice):

    def __init__(self, name, image_name='roiimage', **kw):
        self.call__init__(ImageCounterDevice, name, image_name, **kw)


class LimaCCDs(ImageCounterDevice):

    def __init__(self, name, image_name='video_last_image',
                 image_ct='video_last_image_counter', **kw):
        self.call__init__(ImageCounterDevice, name, image_name, image_ct, **kw)
        self.debug("Prepared to listen image counter (%s) for the %s images"
                   % (self.getImageIDAttrName(), self.getImageAttrNames()))



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""This module contains ANSI color codes"""

from builtins import object

__all__ = ["make_color_table", "NoColors", "TermColors", "HTMLColors"]

__docformat__ = "restructuredtext"


def make_color_table(in_class, use_name=False, fake=False):
    """Build a set of color attributes in a class.
    Helper function for building the TermColors classes."""
    color_templates = (
        ("Black", "0;30"),
        ("Red", "0;31"),
        ("Green", "0;32"),
        ("Brown", "0;33"),
        ("Blue", "0;34"),
        ("Purple", "0;35"),
        ("Cyan", "0;36"),
        ("LightGray", "0;37"),
        ("DarkGray", "1;30"),
        ("LightRed", "1;31"),
        ("LightGreen", "1;32"),
        ("Yellow", "1;33"),
        ("LightBlue", "1;34"),
        ("LightPurple", "1;35"),
        ("LightCyan", "1;36"),
        ("White", "1;37"),)
    if fake:
        for name, value in color_templates:
            setattr(in_class, name, "")
    else:
        if use_name:
            for name, value in color_templates:
                setattr(in_class, name, in_class._base % name)
        else:
            for name, value in color_templates:
                setattr(in_class, name, in_class._base % value)


class NoColors(object):
    NoColor = ''  # for color schemes in color-less terminals.
    Normal = ''   # Reset normal coloring
    _base = ''  # Template for all other colors


class TermColors(object):
    """Color escape sequences.

    This class defines the escape sequences for all the standard (ANSI?)
    colors in terminals. Also defines a NoColor escape which is just the null
    string, suitable for defining 'dummy' color schemes in terminals which get
    confused by color escapes.

    This class should be used as a mixin for building color schemes.

    Basicaly this class is just a copy of IPython.ColorANSI.TermColors class"""

    NoColor = ''  # for color schemes in color-less terminals.
    Normal = '\033[0m'   # Reset normal coloring
    _base = '\033[%sm'  # Template for all other colors


class HTMLColors(object):

    NoColor = ''
    Normal = '</font>'
    _base = '<font color=%s>'

# Build the actual color table as a set of class attributes:
make_color_table(NoColors, fake=True)
make_color_table(TermColors)
make_color_table(HTMLColors, True)



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
Utilities for plugin loading and selection
"""

__all__ = ["selectEntryPoints", "EntryPointAlike"]

import re
import pkg_resources
try:
    from re import fullmatch
except ImportError:  # TODO: remove this when dropping support of py2
    def fullmatch(regex, string, flags=0):
        """
        Emulate python-3.4 re.fullmatch()...  mostly
        See: https://stackoverflow.com/a/30212414
        """
        m = re.match(regex, string, flags=flags)
        if m and m.span()[1] == len(string):
            return m


class EntryPointAlike(object):
    """
    A dummy EntryPoint substitute to be used with :func:`selectEntryPoints`
    for wrapping arbitrary objects and imitate the `name` and `load()` API of
    a :class:`pkg_resources.EntryPoint` instance.

    Pass an object and optionally a name to the constructor to get the wrapped
    `EntryPointAlike` instance. The `repr` of the object is used as the name
    if `name` arg is not provided.
    """
    def __init__(self, obj, name=None):
        self._obj = obj
        if name is None:
            name = repr(obj)
        self.name = name

    def load(self):
        """Returns the wrapped object"""
        return self._obj


def selectEntryPoints(group=None, include=('.*',), exclude=()):
    """
    Selects and prioritizes entry points from an entry point group.

    The entry points are selected using their name.

    The selection is done by regex matching on the names: first the entry
    points whose name matches any of the patterns in `excluded` are discarded;
    then each pattern in `included` is used to select from the names of the
    remaining entry points (from highest to lowest priority).

    In this way, the entry points are selected in the order dictated by
    the `included` pattern list. If a pattern matches several names, these
    will be sorted alphabetically. If a name is matched by several patterns,
    it is only selected by the first pattern.

    For example, if there are the following registered entry point names:
    ['foo1', 'foo2', 'baz1', 'baz2', 'bar1', 'bar2']
    And we use `exclude=("foo2",)` and `include=("bar2", "b.*1", "f.*")` ,
    then the selection will be: ["bar2","bar1","baz1","foo1"]

    Note: apart from regex patterns (strings or compiled) the `include` list
    can also contain :class:`pkg_resources`EntryPoint`-like instances
    (more specifically, an object having `.name` and `.load()` members), in
    which case they are added directly to the selected list. If a member is
    something other than a pattern or an EntryPoint-like object, it will be
    wrapped in an :class:`EntryPointAlike` instance and also included in the
    selection.

    :param group: (str) entry point group name from which the entry points
                  are obtained.
    :param include: (`tuple`). The members of the tuple can either be
                    patterns (both in the form of strings or of
                    :class:`re.Pattern` objects), which will be matched against
                    registered names in group; or EntryPoint-like objects which
                    will be included as they are; or an arbitrary object which
                    will be wrapped as an EntryPoint-like object before being
                    included. Default is `(".*",)`, which matches all
                    registered names in group and the sort is purely
                    alphabetical.
    :param exclude: (`tuple`). Regexp patterns (either `str` or
                    :class:`re.Pattern` objects) matching names to be excluded.
                    Default is `()`, so no entry point is excluded.

    :return: (list of :class:`pkg_resources.EntryPoint`) the selected entry
             points.
    """
    ret = []

    remaining = list(pkg_resources.iter_entry_points(group))

    # filter out the entry points whose name matches a exclude pattern
    for p in exclude:
        remaining = [e for e in remaining if not fullmatch(p, e.name)]

    # sort the remaining entry points alphabetically
    remaining.sort(key=lambda e: e.name)

    # fill the ret list with the entry points whose name matches a pattern in
    # the `include` tuple. The inclusion order follows the order of the
    # patterns in `include` (and alphabetically for a pattern that produces
    # multiple matches)
    for p in include:
        try:
            # check if it is a pattern string or pattern object
            p = re.compile(p)
        except TypeError:
            # if p is not an entry point -like object, create one
            if not hasattr(p, 'name') or not hasattr(p, 'load'):
                p = EntryPointAlike(p)
            # and add it directly
            ret.append(p)
            continue
        # if it is a pattern, match it against remaining entry points
        tmp = remaining
        remaining = []
        for e in tmp:
            if fullmatch(p, e.name):
                ret.append(e)
            else:
                remaining.append(e)
    return ret



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

"""
curvesAppearanceChooserDlg.py:
    A Qt dialog for choosing plot appearance (symbols and lines)
    for a QwtPlot-derived widget (like Taurusplot)
"""

from builtins import range
from taurus.external.qt import Qt, Qwt5
from datetime import datetime
from taurus.qt.qtgui.util.ui import UILoadable


@UILoadable(with_ui='ui')
class CurveStatsDialog(Qt.QDialog):
    """
    A dialog for configuring and displaying statistics from  curves of a plot
    """

    closed = Qt.pyqtSignal()

    statColumns = ('points', 'min', 'max', 'mean', 'std', 'rms')

    def __init__(self, parent=None):
        super(CurveStatsDialog, self).__init__(parent)
        self.loadUi()

        plot = parent
        xIsTime = plot.getXIsTime()
        self.ui.minDTE.setVisible(xIsTime)
        self.ui.maxDTE.setVisible(xIsTime)
        self.ui.minSB.setVisible(not xIsTime)
        self.ui.maxSB.setVisible(not xIsTime)

        self.ui.minSB.setMinimum(float('-inf'))
        self.ui.minSB.setMaximum(float('inf'))
        self.ui.maxSB.setMinimum(float('-inf'))
        self.ui.maxSB.setMaximum(float('inf'))

        icon = Qt.QIcon('actions:system-search.svg')
        self.ui.selectMinPB.setIcon(icon)
        self.ui.selectMaxPB.setIcon(icon)

        self.refreshCurves()

        cbs = (self.ui.npointsStatCB, self.ui.minStatCB, self.ui.maxStatCB,
               self.ui.meanStatCB, self.ui.stdStatCB, self.ui.rmsStatCB)
        self._checkboxToColMap = dict(zip(cbs, range(len(self.statColumns))))

        self.minPicker = Qwt5.QwtPlotPicker(Qwt5.QwtPlot.xBottom,
                                            Qwt5.QwtPlot.yLeft,
                                            Qwt5.QwtPicker.PointSelection,
                                            Qwt5.QwtPicker.VLineRubberBand,
                                            Qwt5.QwtPicker.AlwaysOn, plot.canvas())

        self.maxPicker = Qwt5.QwtPlotPicker(Qwt5.QwtPlot.xBottom,
                                            Qwt5.QwtPlot.yLeft,
                                            Qwt5.QwtPicker.PointSelection,
                                            Qwt5.QwtPicker.VLineRubberBand,
                                            Qwt5.QwtPicker.AlwaysOn, plot.canvas())

        self.minPicker.setEnabled(False)
        self.maxPicker.setEnabled(False)

        # initialize min and max display
        xmin = plot.axisScaleDiv(Qwt5.QwtPlot.xBottom).lowerBound()
        xmax = plot.axisScaleDiv(Qwt5.QwtPlot.xBottom).upperBound()

        self.minMarker = Qwt5.QwtPlotMarker()
        self.minMarker.setLineStyle(Qwt5.QwtPlotMarker.VLine)
        self.minMarker.setLinePen(Qt.QPen(Qt.Qt.green, 3))
        self.minMarker.setXValue(xmin)
        self.minMarker.attach(plot)
        self.minMarker.hide()

        self.maxMarker = Qwt5.QwtPlotMarker()
        self.maxMarker.setLineStyle(Qwt5.QwtPlotMarker.VLine)
        self.maxMarker.setLinePen(Qt.QPen(Qt.Qt.red, 3))
        self.maxMarker.setXValue(xmax)
        self.maxMarker.attach(plot)
        self.maxMarker.hide()

        if xIsTime:
            self.ui.minDTE.setDateTime(self._timestamptToQDateTime(xmin))
            self.ui.maxDTE.setDateTime(self._timestamptToQDateTime(xmax))
        else:
            self.ui.minSB.setValue(xmin)
            self.ui.maxSB.setValue(xmax)

        refreshAction = Qt.QAction(Qt.QIcon.fromTheme(
            'view-refresh'), "Refresh available curves", self.ui.statsTW)
        refreshAction.setShortcut(Qt.Qt.Key_F5)
        refreshAction.triggered.connect(self.refreshCurves)
        self.ui.statsTW.addAction(refreshAction)

        # connections
        for cb in cbs:
            cb.toggled.connect(self.onStatToggled)
        self.ui.calculatePB.clicked.connect(self.onCalculate)
        self.ui.selectMinPB.clicked.connect(self.onSelectMin)
        self.ui.selectMaxPB.clicked.connect(self.onSelectMax)
        self.minPicker.selected.connect(self.minSelected)
        self.maxPicker.selected.connect(self.maxSelected)
        self.ui.minSB.valueChanged.connect(self.onMinChanged)
        self.ui.minDTE.dateTimeChanged.connect(self.onMinChanged)
        self.ui.maxSB.valueChanged.connect(self.onMaxChanged)
        self.ui.maxDTE.dateTimeChanged.connect(self.onMaxChanged)
        self.ui.minCB.toggled.connect(self.onMinChanged)
        self.ui.maxCB.toggled.connect(self.onMaxChanged)

    def _timestamptToQDateTime(self, ts):
        dt = datetime.fromtimestamp(ts)
        return Qt.QDateTime(dt.year, dt.month, dt.day, dt.hour, dt.minute,
                            dt.second, dt.microsecond // 1000)

    def _QDateTimeToTimestamp(self, qdt):
        return qdt.toTime_t() + qdt.time().msec() / 1000.

    def onSelectMin(self):
        '''slot called when the user clicks on the selectMin button'''
        plot = self.parent()
        plot.getZoomers()[0].setEnabled(False)
        plot.canvas().setCursor(Qt.Qt.SplitHCursor)
        self.minPicker.setEnabled(True)
        self.hide()

    def minSelected(self, pos):
        '''slot called when the user has selected a min value from the plot'''
        self.show()
        self.ui.minCB.setChecked(True)
        self.minPicker.setEnabled(False)
        plot = self.parent()
        xmin = pos.x()
        if plot.getXIsTime():
            # this triggers a call to onMinChanged()
            self.ui.minDTE.setDateTime(self._timestamptToQDateTime(xmin))
        else:
            # this triggers a call to onMinChanged()
            self.ui.minSB.setValue(xmin)
        self.restorePlot(keepMarkers=True)

    def onMinChanged(self, x):
        '''slot called when the min value is changed'''
        plot = self.parent()
        if isinstance(x, bool):
            self.minMarker.setVisible(x)
        else:
            if isinstance(x, Qt.QDateTime):
                x = self._QDateTimeToTimestamp(x)
            self.minMarker.setXValue(x)
        plot.replot()

    def onSelectMax(self):
        '''slot called when the user clicks on the selectMax button'''
        self.minPicker.setEnabled(False)
        plot = self.parent()
        plot.getZoomers()[0].setEnabled(False)
        plot.canvas().setCursor(Qt.Qt.SplitHCursor)
        self.maxPicker.setEnabled(True)
        self.hide()

    def maxSelected(self, pos):
        '''slot called when the user has selected a min value from the plot'''
        self.show()
        self.ui.maxCB.setChecked(True)
        self.maxPicker.setEnabled(False)
        plot = self.parent()
        xmax = pos.x()
        if plot.getXIsTime():
            # this triggers a call to onMaxChanged()
            self.ui.maxDTE.setDateTime(self._timestamptToQDateTime(xmax))
        else:
            # this triggers a call to onMaxChanged()
            self.ui.maxSB.setValue(xmax)
        self.restorePlot(keepMarkers=True)

    def onMaxChanged(self, x):
        '''slot called when the max value is changed'''
        plot = self.parent()
        if isinstance(x, bool):
            self.maxMarker.setVisible(x)
        else:
            if isinstance(x, Qt.QDateTime):
                x = self._QDateTimeToTimestamp(x)
            self.maxMarker.setXValue(x)
        plot.replot()

    def onStatToggled(self, checked):
        '''slot called when any of the stat checkboxes is toggled'''
        cb = self.sender()
        col = self._checkboxToColMap[cb]
        if checked:
            self.ui.statsTW.showColumn(col)
        else:
            self.ui.statsTW.hideColumn(col)

    def refreshCurves(self):
        '''resets the table re-reading the curves from the plot'''
        plot = self.parent()
        self.ui.statsTW.clear()
        self.ui.statsTW.setColumnCount(len(self.statColumns))
        self.ui.statsTW.setHorizontalHeaderLabels(self.statColumns)
        self.curveNames = plot.getCurveNamesSorted()
        self.ui.statsTW.setRowCount(len(self.curveNames))
        for row, name in enumerate(self.curveNames):
            title = plot.getCurveTitle(name)
            item = Qt.QTableWidgetItem(title)
            self.ui.statsTW.setVerticalHeaderItem(row, item)

    def getSelectedRows(self):
        '''returns a list of row numbers corresponding to the selected rows of the table'''
        selected = []
        for rg in self.ui.statsTW.selectedRanges():
            for row in range(rg.topRow(), rg.topRow() + rg.rowCount()):
                selected.append(row)
        return selected

    def onCalculate(self):
        '''
        slot called when the calculate button is pressed. Performs the
        calculation of stats in the current limits for the currently selected
        curves (or for all if none selected) and fills the table.
        '''
        plot = self.parent()
        table = self.ui.statsTW

        xmin, xmax = None, None
        if self.ui.minCB.isChecked():
            if plot.getXIsTime():
                xmin = (self.ui.minDTE.dateTime().toTime_t()
                        + self.ui.minDTE.time().msec() / 1000.)
            else:
                xmin = self.ui.minSB.value()
        if self.ui.maxCB.isChecked():
            if plot.getXIsTime():
                xmax = (self.ui.maxDTE.dateTime().toTime_t()
                        + self.ui.maxDTE.time().msec() / 1000.)
            else:
                xmax = self.ui.maxSB.value()
        limits = xmin, xmax

        selectedRows = self.getSelectedRows()
        if len(selectedRows) == 0:
            selectedRows = list(range(len(self.curveNames)))
        selectedCurves = [self.curveNames[i] for i in selectedRows]
        statsdict = plot.getCurveStats(
            limits=limits, curveNames=selectedCurves)

        for row, name in zip(selectedRows, selectedCurves):
            stats = statsdict.get(name, {})

            minxy = stats.get('min')
            if minxy is None:
                text = "---"
            elif plot.xIsTime:
                text = "t=%s \ny=%g" % (datetime.fromtimestamp(
                    minxy[0]).isoformat(), minxy[1])
            else:
                text = "x=%g \ny=%g" % minxy
            table.setItem(row, self.statColumns.index(
                'min'), Qt.QTableWidgetItem(text))

            maxxy = stats.get('max')
            if maxxy is None:
                text = "---"
            elif plot.xIsTime:
                text = "t=%s \ny=%g" % (datetime.fromtimestamp(
                    maxxy[0]).isoformat(), maxxy[1])
            else:
                text = "x=%g \ny=%g" % maxxy
            table.setItem(row, self.statColumns.index(
                'max'), Qt.QTableWidgetItem(text))

            for s in ('points', 'mean', 'std', 'rms'):
                r = stats.get(s)
                if r is None:
                    text = "---"
                else:
                    text = "%g" % r
                table.setItem(row, self.statColumns.index(s),
                              Qt.QTableWidgetItem(text))
        table.resizeColumnsToContents()

    def restorePlot(self, keepMarkers=False):
        '''leaves the parent plot in its original state'''
        plot = self.parent()
        plot.canvas().setCursor(Qt.Qt.CrossCursor)
        plot.getZoomers()[0].setEnabled(plot.getAllowZoomers())
        if not keepMarkers:
            self.minMarker.detach()
            self.maxMarker.detach()
        plot.replot()

    def closeEvent(self, event):
        '''See :meth:`Qwidget.closeEvent`'''
        self.restorePlot()
        self.closed.emit()

    def showEvent(self, event):
        '''See :meth:`Qwidget.showEvent`'''
        plot = self.parent()
        self.minMarker.attach(plot)
        self.maxMarker.attach(plot)
        plot.replot()



# -*- coding: utf-8 -*-
#
# Copyright © 2018- CELLS / ALBA Synchrotron, Bellaterra, Spain
# Copyright © 2014-2015 Colin Duquesnoy
# Copyright © 2009-2018 The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

"""
Provides QtGui classes and functions.
.. warning:: Contrary to qtpy.QtGui, this module exposes the namespace
    available in ``PyQt4.QtGui``.
    See: http://pyqt.sourceforge.net/Docs/PyQt5/pyqt4_differences.html#qtgui-module
"""
import warnings

from . import PYQT5, PYQT4, PYSIDE, PYSIDE2, PythonQtError


if PYQT5:
    from PyQt5.QtGui import *
    # import * from QtWidgets and QtPrintSupport for PyQt4 style compat
    from PyQt5.QtWidgets import *
    from PyQt5.QtPrintSupport import *
elif PYSIDE2:
    from PySide2.QtGui import *
    # import * from QtWidgets and QtPrintSupport for PyQt4 style compat
    from PySide2.QtWidgets import *
    from PySide2.QtPrintSupport import *
elif PYQT4:
    from PyQt4.QtGui import *

elif PYSIDE:
    from PySide.QtGui import *
else:
    raise PythonQtError('No Qt bindings could be found')



# -*- coding: utf-8 -*-

##############################################################################
##
## This file is part of Taurus
##
## http://taurus-scada.org
##
## Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
## Taurus is free software: you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## Taurus is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import sys

import taurus
import unittest


def _import(name):
    __import__(name)
    return sys.modules[name]


class QtTestCase(unittest.TestCase):

    _api_name = None

    def setUp(self):
        taurus.setLogLevel(taurus.Critical)

        self.opt_mods = ("QtDesigner", "QtNetwork", "Qt", "QtSvg",
                         "QtUiTools", "QtWebKit", "Qwt5", "uic")

        # store a "snapshot" of the currently loaded modules
        self._orig_mods = set(sys.modules.keys())

        # this import initializes Qt in case it is not loaded
        from taurus.external.qt import Qt, API_NAME

        self._api_name = API_NAME
        self.__qt = Qt

    def test_qt_base_import(self):
        mods = set(sys.modules.keys())

        other_apis = set(('PyQt5', 'PySide2', 'PyQt4', 'PySide'))
        other_apis.remove(self._api_name)

        # the selected API and the QtCore should be loaded
        self.assertTrue(self._api_name in mods, self._api_name + " not loaded")
        self.assertTrue(self._api_name + ".QtCore" in mods,
                        "QtCore not loaded")

        # the other APIs should *not* be loaded
        for other_api in other_apis:
            self.assertFalse(
                other_api in mods,
                other_api + " loaded in " + self._api_name + " test")

        # the other Qt submodules should *not* be loaded
        for opt_mod in self.opt_mods:
            mod = "{0}.{1}".format(self._api_name, opt_mod)
            self.assertFalse(mod in mods - self._orig_mods, mod + " is loaded")

    def __test_qt_module(self, qt_mod_name):
        """Checks that the given shim is complete"""
        taurus_qt_mod_name = "taurus.external.qt.{0}".format(qt_mod_name)
        orig_qt_mod_name = "{0}.{1}".format(self._api_name, qt_mod_name)
        TaurusQtMod = _import(taurus_qt_mod_name)
        OrigQtMod = _import(orig_qt_mod_name)
        taurus_qt_mod_members = [m for m in dir(TaurusQtMod)
                                 if not m.startswith("_")]
        orig_qt_mod_members = [m for m in dir(OrigQtMod)
                               if not m.startswith("_")]

        for orig_member_name in orig_qt_mod_members:
            self.assertTrue(
                orig_member_name in taurus_qt_mod_members,
                "Taurus {0} does not contain {1}".format(qt_mod_name,
                                                         orig_member_name)
            )

    def test_qt_core(self):
        """Check the QtCore shim"""
        return self.__test_qt_module("QtCore")

    def test_qt_gui(self):
        """Check the QtGui shim"""
        return self.__test_qt_module("QtGui")


def main():
    unittest.main(verbosity=2)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

from __future__ import print_function
from ast import literal_eval


def parse_args(s, strip_pars=False):
    """
    Parse a string representing arguments to a method to return the
    corresponding args and kwargs.

    :param s: string representing arguments to a method
    :param strip_pars: (bool) If True, expect s to include surrounding
          parenthesis
    :return: args, kwargs (a list of positional arguments and a dict of keyword
             arguments)
    """
    s = s.strip()
    if strip_pars:
        # remove parentheses
        if s.startswith('(') and s.endswith(')'):
            s = s[1:-1].strip()
    if not s:
        # empty arg list
        return [], {}
    a = []
    kw = {}
    for e in s.split(','):
        if '=' in e:
            k, v = e.split('=')
            kw[k.strip()] = literal_eval(v.strip())
        else:
            if kw:
                # a non-kwarg found after we already have at least one kwarg
                raise SyntaxError('non-keyword arg after keyword arg')
            a.append(literal_eval(e.strip()))
    return a, kw

if __name__ == "__main__":

    print(parse_args('1, 2, b=3, c=4'))
    print(parse_args(' (1, 2, b=3, c=4 )', strip_pars=True))

    print(parse_args('1, 2, b=3, c=4, 5'))  # <--this should raise a SyntaxError



#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################

import pytest
from taurus.core.util.test.test_plugin import mock_entry_point
from taurus.cli.alt import (
    _load_class_from_group,
    plot_cmd,
    trend_cmd,
    trend2d_cmd,
    image_cmd,
)
from click.testing import CliRunner

runner = CliRunner()


class _MockPlot(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.model = None

    def setModel(self, model):
        self.model = model


def test_load_good_class_from_group():
    """Check get_plot_class with several registered plot alternatives"""
    lines = ["mock{}={}:_MockPlot".format(i, __name__) for i in range(3)]
    mapping = mock_entry_point(lines)
    group_name = list(mapping.keys())[0]
    assert _load_class_from_group(group_name) == (_MockPlot, "mock0")


def test_load_onebad_class_from_group():
    """Check get_plot_class with a bad and a good plot alternatives"""
    lines = ["bad=_unimportablemod:Bad", "good={}:_MockPlot".format(__name__)]
    mapping = mock_entry_point(lines)
    group_name = list(mapping.keys())[0]
    assert _load_class_from_group(group_name) == (_MockPlot, "good")


def test_load_onlybad_class_from_group():
    """Check get_plot_class with no plot alternatives"""
    lines = ["bad=_unimportablemod:Bad"]
    mapping = mock_entry_point(lines)
    group_name = list(mapping.keys())[0]
    with pytest.raises(ImportError) as exc_info:
        _load_class_from_group(group_name)
    assert "bad = _unimportablemod:Bad" in str(exc_info.value)


def test_plot_cmd_options():
    response = runner.invoke(plot_cmd, ["--ls-alt"])
    assert response.exit_code == 0
    assert "Registered alternatives" in response.output

    response = runner.invoke(plot_cmd, ["--use-alt", "_non_existent_"])
    assert "Registered alternatives" in response.output
    assert response.exit_code == 1

    response = runner.invoke(plot_cmd, ["-x", "unsupported"])
    assert "Invalid value" in response.output
    assert response.exit_code == 2

    response = runner.invoke(plot_cmd, ["--config", "_non_existent_"])
    assert "Invalid value" in response.output
    assert "No such file or directory" in response.output
    assert response.exit_code == 2


def test_plot_cmd_help():
    response = runner.invoke(plot_cmd, ["--help"])
    assert "--demo" in response.output
    assert "--ls-alt" in response.output
    assert "--use-alt" in response.output
    assert "--config" in response.output
    assert "--window-name" in response.output
    assert "--x-axis-mode" in response.output
    assert "-x" in response.output
    assert "--help" in response.output
    assert response.exit_code == 0


def test_trend_cmd_options():
    response = runner.invoke(trend_cmd, ["--ls-alt"])
    assert response.exit_code == 0
    assert "Registered alternatives" in response.output

    response = runner.invoke(trend_cmd, ["--use-alt", "_non_existent_"])
    assert "Registered alternatives" in response.output
    assert response.exit_code == 1

    response = runner.invoke(trend_cmd, ["-x", "unsupported"])
    assert "Invalid value" in response.output
    assert response.exit_code == 2

    response = runner.invoke(trend_cmd, ["--config", "_non_existent_"])
    assert "Invalid value" in response.output
    assert "No such file or directory" in response.output
    assert response.exit_code == 2


def test_trend_cmd_help():
    response = runner.invoke(trend_cmd, ["--help"])
    assert "--demo" in response.output
    assert "--ls-alt" in response.output
    assert "--use-alt" in response.output
    assert "--config" in response.output
    assert "--window-name" in response.output
    assert "--x-axis-mode" in response.output
    assert "-x" in response.output
    assert "--buffer" in response.output
    assert "-b" in response.output
    assert "--forced-read" in response.output
    assert "-r" in response.output
    assert "--help" in response.output
    assert response.exit_code == 0


def test_trend2d_cmd_options():
    response = runner.invoke(trend2d_cmd, ["--ls-alt"])
    assert response.exit_code == 0
    assert "Registered alternatives" in response.output

    response = runner.invoke(trend2d_cmd, ["--use-alt", "_non_existent_"])
    assert "Registered alternatives" in response.output
    assert response.exit_code == 1

    response = runner.invoke(trend2d_cmd, ["-x", "unsupported"])
    assert "Invalid value" in response.output
    assert response.exit_code == 2


def test_trend2d_cmd_help():
    response = runner.invoke(trend2d_cmd, ["--help"])
    assert "--demo" in response.output
    assert "--ls-alt" in response.output
    assert "--use-alt" in response.output
    assert "--window-name" in response.output
    assert "--x-axis-mode" in response.output
    assert "-x" in response.output
    assert "--buffer" in response.output
    assert "-b" in response.output
    assert "--help" in response.output
    assert response.exit_code == 0


def test_image_cmd_options():
    response = runner.invoke(image_cmd, ["--ls-alt"])
    assert response.exit_code == 0
    assert "Registered alternatives" in response.output

    response = runner.invoke(image_cmd, ["--use-alt", "_non_existent_"])
    assert "Registered alternatives" in response.output
    assert response.exit_code == 1

    response = runner.invoke(image_cmd, ["-c", "unsupported"])
    assert "Invalid value" in response.output
    assert response.exit_code == 2


def test_image_cmd_help():
    response = runner.invoke(image_cmd, ["--help"])
    assert "--demo" in response.output
    assert "--ls-alt" in response.output
    assert "--use-alt" in response.output
    assert "--window-name" in response.output
    assert "--color-mode" in response.output
    assert "-c" in response.output
    assert "--help" in response.output
    assert response.exit_code == 0
